#
# Copyright (C) 2019 FreeIPA Contributors see COPYING for license
#

from base import BaseTest
from unittest.mock import Mock, patch
from util import capture_results, CAInstance, KRAInstance

from ipahealthcheck.core import config, constants
from ipahealthcheck.ipa.plugin import registry
from ipahealthcheck.ipa.certs import IPARAAgent, IPAKRAAgent

from ipalib import errors
from ipapython.dn import DN
from ipapython.ipaldap import LDAPClient, LDAPEntry
from ipaplatform.paths import paths

from ldap import OPT_X_SASL_SSF_MIN


class IPACertificate:
    def __init__(self, serial_number=1):
        self.subject = 'CN=RA AGENT'
        self.issuer = 'CN=ISSUER'
        self.serial_number = serial_number


class mock_ldap:
    SCOPE_BASE = 1
    SCOPE_ONELEVEL = 2
    SCOPE_SUBTREE = 4

    def __init__(self, ldapentry):
        """Initialize the results that we will return from get_entries"""
        self.results = ldapentry

    def get_entries(self, base_dn, scope=SCOPE_SUBTREE, filter=None,
                    attrs_list=None, get_effective_rights=False, **kwargs):
        if self.results is None:
            raise errors.NotFound(reason='test')
        return self.results


class mock_ldap_conn:
    def set_option(self, option, invalue):
        pass

    def get_option(self, option):
        if option == OPT_X_SASL_SSF_MIN:
            return 256

        return None

    def search_s(self, base, scope, filterstr=None,
                 attrlist=None, attrsonly=0):
        return tuple()


class TestNSSAgent(BaseTest):
    cert = IPACertificate()
    patches = {
        'ldap.initialize':
        Mock(return_value=mock_ldap_conn()),
        'ipaserver.install.cainstance.CAInstance':
        Mock(return_value=CAInstance()),
        'ipalib.x509.load_certificate_from_file':
        Mock(return_value=cert),
    }

    def test_nss_agent_ok(self):

        attrs = dict(
            description=['2;1;CN=ISSUER;CN=RA AGENT'],
            usercertificate=[self.cert],
        )
        fake_conn = LDAPClient('ldap://localhost', no_schema=True)
        ldapentry = LDAPEntry(fake_conn, DN('uid=ipara,ou=people,o=ipaca'))
        for attr, values in attrs.items():
            ldapentry[attr] = values

        framework = object()
        registry.initialize(framework, config.Config())
        f = IPARAAgent(registry)

        f.conn = mock_ldap([ldapentry])
        self.results = capture_results(f)

        assert len(self.results) == 1

        result = self.results.results[0]
        assert result.result == constants.SUCCESS
        assert result.source == 'ipahealthcheck.ipa.certs'
        assert result.check == 'IPARAAgent'

    def test_nss_agent_no_description(self):

        attrs = dict(
            usercertificate=[self.cert],
        )
        fake_conn = LDAPClient('ldap://localhost', no_schema=True)
        ldapentry = LDAPEntry(fake_conn, DN('uid=ipara,ou=people,o=ipaca'))
        for attr, values in attrs.items():
            ldapentry[attr] = values

        framework = object()
        registry.initialize(framework, config.Config())
        f = IPARAAgent(registry)

        f.conn = mock_ldap([ldapentry])
        self.results = capture_results(f)
        result = self.results.results[0]

        assert result.result == constants.ERROR
        assert 'description' in result.kw.get('msg')

    @patch('ipalib.x509.load_certificate_from_file')
    def test_nss_agent_load_failure(self, mock_load_cert):

        mock_load_cert.side_effect = IOError('test')

        framework = object()
        registry.initialize(framework, config.Config())
        f = IPARAAgent(registry)

        self.results = capture_results(f)
        result = self.results.results[0]

        assert result.result == constants.ERROR
        assert result.kw.get('error') == 'test'

    def test_nss_agent_no_entry_found(self):

        framework = object()
        registry.initialize(framework, config.Config())
        f = IPARAAgent(registry)

        f.conn = mock_ldap(None)  # None == NotFound
        self.results = capture_results(f)
        result = self.results.results[0]

        assert result.result == constants.ERROR
        assert result.kw.get('msg') == 'RA agent not found in LDAP'

    def test_nss_agent_too_many(self):

        attrs = dict(
            description=['2;1;CN=ISSUER;CN=RA AGENT'],
            usercertificate=[self.cert],
        )
        fake_conn = LDAPClient('ldap://localhost', no_schema=True)
        ldapentry = LDAPEntry(fake_conn, DN('uid=ipara,ou=people,o=ipaca'))
        for attr, values in attrs.items():
            ldapentry[attr] = values

        ldapentry2 = LDAPEntry(fake_conn, DN('uid=ipara2,ou=people,o=ipaca'))
        for attr, values in attrs.items():
            ldapentry[attr] = values

        framework = object()
        registry.initialize(framework, config.Config())
        f = IPARAAgent(registry)

        f.conn = mock_ldap([ldapentry, ldapentry2])
        self.results = capture_results(f)
        result = self.results.results[0]

        assert result.result == constants.ERROR
        assert result.kw.get('found') == 2

    def test_nss_agent_nonmatching_cert(self):

        cert2 = IPACertificate(2)

        attrs = dict(
            description=['2;1;CN=ISSUER;CN=RA AGENT'],
            usercertificate=[cert2],
        )
        fake_conn = LDAPClient('ldap://localhost', no_schema=True)
        ldapentry = LDAPEntry(fake_conn, DN('uid=ipara,ou=people,o=ipaca'))
        for attr, values in attrs.items():
            ldapentry[attr] = values

        framework = object()
        registry.initialize(framework, config.Config())
        f = IPARAAgent(registry)

        f.conn = mock_ldap([ldapentry])
        self.results = capture_results(f)
        result = self.results.results[0]

        assert result.result == constants.ERROR
        assert result.kw.get('certfile') == paths.RA_AGENT_PEM
        assert result.kw.get('dn') == 'uid=ipara,ou=people,o=ipaca'

    def test_nss_agent_multiple_certs(self):

        cert2 = IPACertificate(2)

        attrs = dict(
            description=['2;1;CN=ISSUER;CN=RA AGENT'],
            usercertificate=[cert2, self.cert],
        )
        fake_conn = LDAPClient('ldap://localhost', no_schema=True)
        ldapentry = LDAPEntry(fake_conn, DN('uid=ipara,ou=people,o=ipaca'))
        for attr, values in attrs.items():
            ldapentry[attr] = values

        framework = object()
        registry.initialize(framework, config.Config)
        f = IPARAAgent(registry)

        f.conn = mock_ldap([ldapentry])
        self.results = capture_results(f)

        assert len(self.results) == 1

        result = self.results.results[0]
        assert result.result == constants.SUCCESS
        assert result.source == 'ipahealthcheck.ipa.certs'
        assert result.check == 'IPARAAgent'

    def test_nss_agent_mixed_case_cert(self):
        attrs = dict(
            description=['2;1;cn=ISSUER;cn=RA AGENT'],
            usercertificate=[self.cert],
        )
        fake_conn = LDAPClient('ldap://localhost', no_schema=True)
        ldapentry = LDAPEntry(fake_conn, DN('uid=ipara,ou=people,o=ipaca'))
        for attr, values in attrs.items():
            ldapentry[attr] = values

        framework = object()
        registry.initialize(framework, config.Config)
        f = IPARAAgent(registry)

        f.conn = mock_ldap([ldapentry])
        self.results = capture_results(f)

        assert len(self.results) == 1

        result = self.results.results[0]
        assert result.result == constants.SUCCESS
        assert result.source == 'ipahealthcheck.ipa.certs'
        assert result.check == 'IPARAAgent'


class TestKRAAgent(BaseTest):
    cert = IPACertificate()
    patches = {
        'ldap.initialize':
        Mock(return_value=mock_ldap_conn()),
        'ipaserver.install.krainstance.KRAInstance':
        Mock(return_value=KRAInstance()),
        'ipalib.x509.load_certificate_from_file':
        Mock(return_value=cert),
    }

    def test_kra_agent_ok(self):

        attrs = dict(
            description=['2;1;CN=ISSUER;CN=RA AGENT'],
            usercertificate=[self.cert],
        )
        fake_conn = LDAPClient('ldap://localhost', no_schema=True)
        ldapentry = LDAPEntry(fake_conn,
                              DN('uid=ipakra,ou=people,o=kra,o=ipaca'))
        for attr, values in attrs.items():
            ldapentry[attr] = values

        framework = object()
        registry.initialize(framework, config.Config())
        f = IPAKRAAgent(registry)

        f.conn = mock_ldap([ldapentry])
        self.results = capture_results(f)

        assert len(self.results) == 1

        result = self.results.results[0]
        assert result.result == constants.SUCCESS
        assert result.source == 'ipahealthcheck.ipa.certs'
        assert result.check == 'IPAKRAAgent'

    def test_kra_agent_no_description(self):

        attrs = dict(
            usercertificate=[self.cert],
        )
        fake_conn = LDAPClient('ldap://localhost', no_schema=True)
        ldapentry = LDAPEntry(fake_conn,
                              DN('uid=ipakra,ou=people,o=kra,o=ipaca'))
        for attr, values in attrs.items():
            ldapentry[attr] = values

        framework = object()
        registry.initialize(framework, config.Config())
        f = IPAKRAAgent(registry)

        f.conn = mock_ldap([ldapentry])
        self.results = capture_results(f)
        result = self.results.results[0]

        assert result.result == constants.ERROR
        assert 'description' in result.kw.get('msg')

    @patch('ipalib.x509.load_certificate_from_file')
    def test_kra_agent_load_failure(self, mock_load_cert):

        mock_load_cert.side_effect = IOError('test')

        framework = object()
        registry.initialize(framework, config.Config())
        f = IPAKRAAgent(registry)

        self.results = capture_results(f)
        result = self.results.results[0]

        assert result.result == constants.ERROR
        assert result.kw.get('error') == 'test'

    def test_kra_agent_no_entry_found(self):

        framework = object()
        registry.initialize(framework, config.Config())
        f = IPAKRAAgent(registry)

        f.conn = mock_ldap(None)  # None == NotFound
        self.results = capture_results(f)
        result = self.results.results[0]

        assert result.result == constants.ERROR
        assert result.kw.get('msg') == 'KRA agent not found in LDAP'

    def test_kra_agent_too_many(self):

        attrs = dict(
            description=['2;1;CN=ISSUER;CN=RA AGENT'],
            usercertificate=[self.cert],
        )
        fake_conn = LDAPClient('ldap://localhost', no_schema=True)
        ldapentry = LDAPEntry(fake_conn,
                              DN('uid=ipakra,ou=people,o=kra,o=ipaca'))
        for attr, values in attrs.items():
            ldapentry[attr] = values

        ldapentry2 = LDAPEntry(fake_conn,
                               DN('uid=ipakra,ou=people,o=kra,o=ipaca'))
        for attr, values in attrs.items():
            ldapentry[attr] = values

        framework = object()
        registry.initialize(framework, config.Config())
        f = IPAKRAAgent(registry)

        f.conn = mock_ldap([ldapentry, ldapentry2])
        self.results = capture_results(f)
        result = self.results.results[0]

        assert result.result == constants.ERROR
        assert result.kw.get('found') == 2

    def test_kra_agent_nonmatching_cert(self):

        cert2 = IPACertificate(2)

        attrs = dict(
            description=['2;1;CN=ISSUER;CN=RA AGENT'],
            usercertificate=[cert2],
        )
        fake_conn = LDAPClient('ldap://localhost', no_schema=True)
        ldapentry = LDAPEntry(fake_conn,
                              DN('uid=ipakra,ou=people,o=kra,o=ipaca'))
        for attr, values in attrs.items():
            ldapentry[attr] = values

        framework = object()
        registry.initialize(framework, config.Config())
        f = IPAKRAAgent(registry)

        f.conn = mock_ldap([ldapentry])
        self.results = capture_results(f)
        result = self.results.results[0]

        assert result.result == constants.ERROR
        assert result.kw.get('certfile') == paths.RA_AGENT_PEM
        assert result.kw.get('dn') == 'uid=ipakra,ou=people,o=kra,o=ipaca'

    def test_kra_agent_multiple_certs(self):

        cert2 = IPACertificate(2)

        attrs = dict(
            description=['2;1;CN=ISSUER;CN=RA AGENT'],
            usercertificate=[cert2, self.cert],
        )
        fake_conn = LDAPClient('ldap://localhost', no_schema=True)
        ldapentry = LDAPEntry(fake_conn,
                              DN('uid=ipakra,ou=people,o=kra,o=ipaca'))
        for attr, values in attrs.items():
            ldapentry[attr] = values

        framework = object()
        registry.initialize(framework, config.Config)
        f = IPAKRAAgent(registry)

        f.conn = mock_ldap([ldapentry])
        self.results = capture_results(f)

        assert len(self.results) == 1

        result = self.results.results[0]
        assert result.result == constants.SUCCESS
        assert result.source == 'ipahealthcheck.ipa.certs'
        assert result.check == 'IPAKRAAgent'



#
# Copyright (C) 2019 FreeIPA Contributors see COPYING for license
#

# Error reporting result
SUCCESS = 0
WARNING = 10
ERROR = 20
CRITICAL = 30

_levelToName = {
    SUCCESS: 'SUCCESS',
    WARNING: 'WARNING',
    ERROR: 'ERROR',
    CRITICAL: 'CRITICAL',
}

_nameToLevel = {
    'SUCCESS': SUCCESS,
    'WARNING': WARNING,
    'ERROR': ERROR,
    'CRITICAL': CRITICAL,
}


def getLevelName(level):
    """
    Translate between level constants and their textual mappings.

    If the level is one of the predefined levels then returns the
    corresponding string.

    If a numeric value corresponding to one of the defined levels
    is passed in instead the corresponding string representation is
    returned.
    """
    name = _levelToName.get(level) or _nameToLevel.get(level)
    if name is not None:
        return name

    return level


def getLevel(name):
    """
    Translate between level text and their numeric constants

    If the level is one of the predefined levels then returns the
    corresponding number.
    """
    level = _nameToLevel.get(name)
    if level is not None:
        return level

    return name


CONFIG_FILE = '/etc/ipahealthcheck/ipahealthcheck.conf'
CONFIG_SECTION = 'default'
EXCLUDE_SECTION = 'excludes'

DEFAULT_TIMEOUT = 10

DEFAULT_CONFIG = {
    'cert_expiration_days': 28,
    'timeout': DEFAULT_TIMEOUT,
}



#
# Copyright (C) 2019 FreeIPA Contributors see COPYING for license
#

from util import capture_results
from base import BaseTest

from ipahealthcheck.ipa.plugin import registry
from ipahealthcheck.meta.services import httpd
from ipahealthcheck.core import config


class TestServices(BaseTest):
    def test_simple_service(self):
        """
        Test a service. It was chosen at random.

        The purpose of this test is to exercise the service check
        code path and not to confirm that a particular service is
        running.
        """
        framework = object()
        registry.initialize(framework, config.Config)
        f = httpd(registry)

        self.results = capture_results(f)

        assert len(self.results) == 1



#
# Copyright (C) 2019 FreeIPA Contributors see COPYING for license
#

from ipahealthcheck.core import constants
from ipahealthcheck.core.plugin import Result, duration
from ipahealthcheck.ipa.plugin import IPAPlugin, registry
from ipalib import api


@registry
class IPAMetaCheck(IPAPlugin):
    """Return meta data for the IPA installation"""
    requires = ('dirsrv',)

    @duration
    def check(self):
        try:
            result = api.Command.server_find(pkey_only=True)
        except Exception as e:
            yield Result(self, constants.ERROR,
                         msg='server-show failed, %s' % e)
        else:
            masters = []
            for server in result['result']:
                masters.append(server['cn'][0])
            yield Result(self, constants.SUCCESS,
                         masters=masters)



#
# Copyright (C) 2019 FreeIPA Contributors see COPYING for license
#

from util import capture_results
from util import m_api
from base import BaseTest

from ipahealthcheck.core import config, constants
from ipahealthcheck.ipa.plugin import registry
from ipahealthcheck.ipa.topology import IPATopologyDomainCheck


class TestTopology(BaseTest):
    def test_topology_ok(self):
        m_api.Command.topologysuffix_verify.side_effect = [
            {
                u'result': {
                    u"in_order": True,
                }
            },
            {
                u'result': {
                    u"in_order": True,
                }
            },
        ]
        m_api.Command.ca_is_enabled.return_value = {'result': True}

        framework = object()
        registry.initialize(framework, config.Config)
        f = IPATopologyDomainCheck(registry)

        self.results = capture_results(f)

        assert len(self.results) == 2

        for result in self.results.results:
            assert result.result == constants.SUCCESS
            assert result.source == 'ipahealthcheck.ipa.topology'
            assert result.check == 'IPATopologyDomainCheck'

    def test_topology_domain_bad(self):
        m_api.Command.topologysuffix_verify.side_effect = [
            {
                u'result': {
                    u"connect_errors": [
                        [
                            u"ipa.example.test",
                            [u"ipa.example.test"],
                            [u"replica2.example.test"]
                        ],
                        [
                            u"replica2.example.test",
                            [u"replica2.example.test"],
                            [u"ipa.example.test"]
                        ]
                    ],
                    u"in_order": False,
                    u"max_agmts": 4,
                    u"max_agmts_errors": []
                }
            },
            {
                u'result': {
                    u"in_order": True,
                }
            },
        ]
        m_api.Command.ca_is_enabled.return_value = {'result': True}

        framework = object()
        registry.initialize(framework, config.Config)
        f = IPATopologyDomainCheck(registry)

        self.results = capture_results(f)

        assert len(self.results) == 3

        for result in self.results.results:
            assert result.source == 'ipahealthcheck.ipa.topology'
            assert result.check == 'IPATopologyDomainCheck'

        # The first two results are failures in the domain suffix, the
        # third is a success in the ca suffix.
        result = self.results.results[0]
        assert result.result == constants.ERROR
        assert result.kw.get('key') == 'ipa.example.test'
        assert result.kw.get('replicas') == ['replica2.example.test']
        assert result.kw.get('suffix') == 'domain'
        assert result.kw.get('type') == 'connect'
        assert 'can\'t contact servers' in result.kw.get('msg')

        result = self.results.results[1]
        assert result.result == constants.ERROR
        assert result.kw.get('key') == 'replica2.example.test'
        assert result.kw.get('replicas') == ['ipa.example.test']
        assert result.kw.get('suffix') == 'domain'
        assert result.kw.get('type') == 'connect'
        assert 'can\'t contact servers' in result.kw.get('msg')

        result = self.results.results[2]
        assert result.result == constants.SUCCESS
        assert result.kw.get('suffix') == 'ca'

    def test_topology_ca_bad(self):
        m_api.Command.topologysuffix_verify.side_effect = [
            {
                u'result': {
                    u"in_order": True,
                }
            },
            {
                u'result': {
                    u"connect_errors": [
                        [
                            u"ipa.example.test",
                            [u"ipa.example.test"],
                            [u"replica2.example.test"]
                        ],
                        [
                            u"replica2.example.test",
                            [u"replica2.example.test"],
                            [u"ipa.example.test"]
                        ]
                    ],
                    u"in_order": False,
                    u"max_agmts": 4,
                    u"max_agmts_errors": []
                }
            },
        ]
        m_api.Command.ca_is_enabled.return_value = {'result': True}

        framework = object()
        registry.initialize(framework, config.Config)
        f = IPATopologyDomainCheck(registry)

        self.results = capture_results(f)

        assert len(self.results) == 3

        for result in self.results.results:
            assert result.source == 'ipahealthcheck.ipa.topology'
            assert result.check == 'IPATopologyDomainCheck'

        # The first result is ok (domain) and the last two are failures
        # (ca)
        result = self.results.results[0]
        assert result.result == constants.SUCCESS
        assert result.kw.get('suffix') == 'domain'

        result = self.results.results[1]
        assert result.result == constants.ERROR
        assert result.kw.get('key') == 'ipa.example.test'
        assert result.kw.get('replicas') == ['replica2.example.test']
        assert result.kw.get('suffix') == 'ca'
        assert result.kw.get('type') == 'connect'
        assert 'can\'t contact servers' in result.kw.get('msg')

        result = self.results.results[2]
        assert result.result == constants.ERROR
        assert result.kw.get('key') == 'replica2.example.test'
        assert result.kw.get('replicas') == ['ipa.example.test']
        assert result.kw.get('suffix') == 'ca'
        assert result.kw.get('type') == 'connect'
        assert 'can\'t contact servers' in result.kw.get('msg')

    def test_topology_domain_max_agmts(self):
        m_api.Command.topologysuffix_verify.side_effect = [
            {
                u'result': {
                    u"connect_errors": [],
                    u"in_order": False,
                    u"max_agmts": 1,
                    u"max_agmts_errors": [
                        [
                            u"ipa.example.test",
                            [u"replica2.example.test"],
                        ],
                    ],
                }
            },
            {
                u'result': {
                    u"in_order": True,
                }
            },
        ]
        m_api.Command.ca_is_enabled.return_value = {'result': True}

        framework = object()
        registry.initialize(framework, config.Config)
        f = IPATopologyDomainCheck(registry)

        self.results = capture_results(f)

        assert len(self.results) == 2

        for result in self.results.results:
            assert result.source == 'ipahealthcheck.ipa.topology'
            assert result.check == 'IPATopologyDomainCheck'

        result = self.results.results[0]
        assert result.result == constants.ERROR
        assert result.kw.get('key') == 'ipa.example.test'
        assert result.kw.get('replicas') == ['replica2.example.test']
        assert result.kw.get('suffix') == 'domain'
        assert result.kw.get('type') == 'max'
        assert 'recommended max' in result.kw.get('msg')

        result = self.results.results[1]
        assert result.result == constants.SUCCESS
        assert result.kw.get('suffix') == 'ca'



#
# Copyright (C) 2019 FreeIPA Contributors see COPYING for license
#

import json
from ipahealthcheck.core.output import OutputRegistry, Output


output_registry = OutputRegistry()

class ClusterOutput(Output):
    """Base class for writing/display output of cluster results

       severity doesn't apply in this case so exclude those.
    """
    def __init__(self, options):
        self.filename = options.output_file

    def strip_output(self, results):
        """Nothing to strip out"""
        return list(results.output())

    def generate(self, data):
        raise NotImplementedError


@output_registry
class Ansible(ClusterOutput):
    """Output information JSON format for consumption by Ansible

       Required keywords in a Result:
       name - unique identifier for the return value

       One of these is required:
       value - the return value. Type? I dunno yet
       error - if an error was returned
    """

    options = (
        ('--indent', dict(dest='indent', type=int, default=2,
         help='Indention level of JSON output')),
    )

    def __init__(self, options):
        super().__init__(options)
        self.indent = options.indent

    def generate(self, data):
        output = []
        for line in data:
            kw = line.get('kw')
            name = kw.get('name')
            value = kw.get('value')
            error = kw.get('error')

            if value and error:
                value = '%s: %s' % (error, value)
            elif error:
                value = error

            rval = {'%s' % name: value}
            output.append(rval)

        output = json.dumps(output, indent=self.indent)
        if self.filename is None:
            output += '\n'

        return output



#
# Copyright (C) 2019 FreeIPA Contributors see COPYING for license
#

from base import BaseTest
from unittest.mock import Mock, patch
from util import capture_results, CAInstance
from ipahealthcheck.core import config, constants
from ipahealthcheck.ipa.plugin import registry
from ipahealthcheck.ipa.certs import IPAOpenSSLChainValidation

from ipapython.ipautil import _RunResult


class TestOpenSSLValidation(BaseTest):
    patches = {
        'ipaserver.install.cainstance.CAInstance':
        Mock(return_value=CAInstance()),
    }

    @patch('ipapython.ipautil.run')
    def test_openssl_validation_ok(self, mock_run):
        def run(args, raiseonerr=True):
            result = _RunResult('', '', 0)
            result.raw_output = bytes(
                '{}: OK'.format(args[-1]).encode('utf-8'))
            result.raw_error_output = b''
            return result

        mock_run.side_effect = run

        framework = object()
        registry.initialize(framework, config.Config)
        f = IPAOpenSSLChainValidation(registry)

        self.results = capture_results(f)

        assert len(self.results) == 2

        for result in self.results.results:
            assert result.result == constants.SUCCESS
            assert result.source == 'ipahealthcheck.ipa.certs'
            assert result.check == 'IPAOpenSSLChainValidation'

    @patch('ipapython.ipautil.run')
    def test_openssl_validation_bad(self, mock_run):
        def run(args, raiseonerr=True):
            result = _RunResult('', '', 2)
            result.raw_output = bytes(
                'O = EXAMPLE.TEST, CN = ipa.example.test\n'
                'error 20 at 0 depth lookup: unable to get local issuer '
                'certificate\nerror {}: verification failed'.format(args[-1])
                .encode('utf-8'))
            result.raw_error_output = b''
            result.error_log = ''
            return result

        mock_run.side_effect = run

        framework = object()
        registry.initialize(framework, config.Config)
        f = IPAOpenSSLChainValidation(registry)

        self.results = capture_results(f)

        assert len(self.results) == 2

        for result in self.results.results:
            assert result.result == constants.ERROR
            assert result.source == 'ipahealthcheck.ipa.certs'
            assert result.check == 'IPAOpenSSLChainValidation'
            assert 'failed' in result.kw.get('msg')



#
# Copyright (C) 2019 FreeIPA Contributors see COPYING for license
#

from datetime import datetime
import json
import sys
from ipahealthcheck.core.constants import _nameToLevel, SUCCESS
from ipahealthcheck.core.plugin import Registry


class OutputRegistry(Registry):
    pass


output_registry = OutputRegistry()


class Output:
    """Base class for writing/displaying the output of results

       options is a tuple of argparse options that can add
       class-specific options for output.

       Output will be typically generated like:
       >>> output = JSON(options)
       >>> output.render(results)

       render() will:
       1. Strip out any SUCCESS if requested (strip_output)
       2. Generate a string to be written (generate)
       3. Write to the requested file or stdout (write_file)

       stdout == sys.stdout by default.

       An Output class only needs to implement the generate() method
       which will render the results into a string for writing.
    """
    def __init__(self, options):
        self.filename = options.output_file

        # Non-required options in the framework, set logical defaults to
        # pre 0.6 behavior with everything reported.
        self.severity = None
        self.failures_only = False
        self.all = True

        if 'failures_only' in options:
            self.failures_only = options.failures_only
        if 'all' in options:
            self.all = options.all
        if 'severity' in options:
            self.severity = options.severity

    def render(self, results):
        """Process the results into output"""
        output = self.strip_output(results)
        output = self.generate(output)
        self.write_file(output)

    def write_file(self, output):
        """Write the output to a file or sys.stdout"""
        if self.filename:
            with open(self.filename, 'w') as fd:
                fd.write(output)
        else:
            sys.stdout.write(output)

    def strip_output(self, results):
        """Strip out SUCCESS results if --failures-only or
           --severity was used

           Returns a list of result values.
        """
        output = []
        for line in results.output():
            result = line.get('result')
            if _nameToLevel.get(result) == SUCCESS:
                if self.failures_only:
                    continue
                if (not self.all and
                    self.filename is None and
                    not (self.failures_only is False and
                         not sys.stdin.isatty())):
                    continue
            if self.severity is not None and result not in self.severity:
                continue
            output.append(line)

        return output

    def generate(self, data):
        """Convert the output to the desired format, ready for writing

           This is the only method an output plugin is required to
           provide. The return value should be in ready-to-write format.

           Returns a string.
        """
        raise NotImplementedError


@output_registry
class JSON(Output):
    """Output information in JSON format"""

    options = (
        ('--indent', dict(dest='indent', type=int, default=2,
         help='Indention level of JSON output')),
    )

    def __init__(self, options):
        super().__init__(options)
        self.indent = options.indent

    def generate(self, data):
        output = json.dumps(data, indent=self.indent)
        if self.filename is None:
            output += '\n'

        return output


@output_registry
class Human(Output):
    """Display output in a more human-friendly way"""
    options = ()

    def generate(self, data):
        if not data:
            return "No issues found.\n"
        output = ''
        for line in data:
            kw = line.get('kw')
            result = line.get('result')
            source = line.get('source')
            check = line.get('check')
            outline = '%s: %s.%s' % (result, source, check)
            if 'key' in kw:
                outline += '.%s' % kw.get('key')
            if 'msg' in kw:
                msg = kw.get('msg')
                err = msg.format(**kw)
                outline += ': %s' % err
            elif 'exception' in kw:
                outline += ': %s' % kw.get('exception')
            output += outline + '\n'

        return output


@output_registry
class Prometheus(Output):
    """Render results as Prometheus text metric exposition format"""

    options = (
        ('--metric-prefix', dict(dest='metric_prefix', default='ipa',
         help='Metric name prefix')),
    )

    def __init__(self, options):
        super().__init__(options)
        self.metric_prefix = options.metric_prefix

    def generate(self, data):
        if not data:
            return '\n'

        crt = {}
        svc = {}
        chk = {}
        for line in data:
            kw = line.get('kw')
            result = line.get('result')
            source = line.get('source')
            check = line.get('check')

            if result in chk:
                chk[result] += 1
            else:
                chk[result] = 1

            if source == 'ipahealthcheck.meta.services':
                state = 1.0 if _nameToLevel.get(result) == SUCCESS else 0.0
                svc[check] = state
            elif (source == 'ipahealthcheck.ipa.certs' and
                  check == "IPACertmongerExpirationCheck"):

                # only unsuccessful checks carry the expiration information
                if 'key' in kw and 'expiration_date' in kw:
                    expiration = datetime.strptime(kw['expiration_date'],
                                                   '%Y%m%d%H%M%SZ')
                    crt[kw['key']] = expiration.timestamp()

        metrics = []
        self.generate_check_metrics(metrics, chk)
        self.generate_service_metrics(metrics, svc)
        self.generate_certificate_metrics(metrics, crt)

        metrics.append('')
        return "\n".join(metrics)

    def generate_check_metrics(self, out, data):
        if not data:
            return

        metric_name = 'healthcheck'
        out.append('HELP %s_%s %s' %
                   (self.metric_prefix, metric_name,
                    'Number of healthchecks with a certain result'))
        out.append('TYPE %s_%s gauge' %
                   (self.metric_prefix, metric_name))
        for check, quantity in data.items():
            out.append('%s_%s{result="%s"} %.1f' %
                       (self.metric_prefix, metric_name,
                        check, quantity))

    def generate_service_metrics(self, out, data):
        if not data:
            return

        metric_name = 'service_state'
        out.append('HELP %s_%s %s' %
                   (self.metric_prefix, metric_name,
                    'State of the services monitored by IPA healthcheck'))
        out.append('TYPE %s_%s gauge' %
                   (self.metric_prefix, metric_name))
        for service, state in data.items():
            out.append('%s_%s{service="%s"} %.1f' %
                       (self.metric_prefix, metric_name,
                        service, state))

    def generate_certificate_metrics(self, out, data):
        if not data:
            return

        metric_name = 'cert_expiration'
        out.append('HELP %s_%s %s' %
                   (self.metric_prefix, metric_name,
                    'Expiration date of certificates in warning/error state'))
        out.append('TYPE %s_%s gauge' %
                   (self.metric_prefix, metric_name))
        for certificate, timestamp in data.items():
            out.append('%s_%s{certificate_request_id="%s"} %.9e' %
                       (self.metric_prefix, metric_name,
                        certificate, timestamp))



#
# Copyright (C) 2019 FreeIPA Contributors see COPYING for license
#

from base import BaseTest
from unittest.mock import Mock, patch
from util import capture_results, CAInstance
from ipapython.ipautil import _RunResult

from ipahealthcheck.core import config, constants
from ipahealthcheck.ipa.plugin import registry
from ipahealthcheck.ipa.certs import IPANSSChainValidation


class DsInstance:
    def get_server_cert_nickname(self, serverid):
        return 'Server-Cert'


class TestNSSValidation(BaseTest):
    patches = {
        'ipahealthcheck.ipa.certs.get_dogtag_cert_password':
        Mock(return_value='foo'),
        'ipaserver.install.dsinstance.DsInstance':
        Mock(return_value=DsInstance()),
    }

    @patch('ipaserver.install.cainstance.CAInstance')
    @patch('ipapython.ipautil.run')
    def test_nss_validation_ok(self, mock_run, mock_cainstance):
        def run(args, raiseonerr=True):
            result = _RunResult('', '', 0)
            result.raw_output = b'certutil: certificate is valid\n'
            result.raw_error_output = b''
            return result

        mock_run.side_effect = run
        mock_cainstance.return_value = CAInstance()

        framework = object()
        registry.initialize(framework, config.Config)
        f = IPANSSChainValidation(registry)

        self.results = capture_results(f)

        assert len(self.results) == 2

        for result in self.results.results:
            assert result.result == constants.SUCCESS
            assert result.source == 'ipahealthcheck.ipa.certs'
            assert result.check == 'IPANSSChainValidation'

    @patch('ipaserver.install.cainstance.CAInstance')
    @patch('ipapython.ipautil.run')
    def test_nss_validation_bad(self, mock_run, mock_cainstance):
        def run(args, raiseonerr=True):
            result = _RunResult('', '', 255)
            result.raw_output = str.encode(
                'certutil: certificate is invalid: Peer\'s certificate issuer '
                'has been marked as not trusted by the user.'
            )
            result.raw_error_output = b''
            result.error_log = ''
            return result

        mock_run.side_effect = run
        mock_cainstance.return_value = CAInstance()

        framework = object()
        registry.initialize(framework, config.Config)
        f = IPANSSChainValidation(registry)

        self.results = capture_results(f)

        assert len(self.results) == 2

        for result in self.results.results:
            assert result.result == constants.ERROR
            assert result.source == 'ipahealthcheck.ipa.certs'
            assert result.check == 'IPANSSChainValidation'

    @patch('ipaserver.install.cainstance.CAInstance')
    @patch('ipapython.ipautil.run')
    def test_nss_validation_ok_no_ca(self, mock_run, mock_cainstance):
        """Test with the CA marked as not configured so there should only
           be a DS certificate to check.
        """
        def run(args, raiseonerr=True):
            result = _RunResult('', '', 0)
            result.raw_output = b'certutil: certificate is valid\n'
            result.raw_error_output = b''
            return result

        mock_run.side_effect = run
        mock_cainstance.return_value = CAInstance(False)

        framework = object()
        registry.initialize(framework, config.Config)
        f = IPANSSChainValidation(registry)

        self.results = capture_results(f)

        assert len(self.results) == 1

        for result in self.results.results:
            assert result.result == constants.SUCCESS
            assert result.source == 'ipahealthcheck.ipa.certs'
            assert result.check == 'IPANSSChainValidation'
            assert 'slapd-' in result.kw.get('key')



#
# Copyright (C) 2021 FreeIPA Contributors see COPYING for license
#

import logging
import lxml.etree
import re

from ipahealthcheck.ipa.plugin import IPAPlugin, registry
from ipahealthcheck.core.plugin import Result
from ipahealthcheck.core.plugin import duration
from ipahealthcheck.core import constants

from ipaplatform.paths import paths


logger = logging.getLogger()


def read_ipa_pki_proxy():
    """Read the IPA Proxy configuration file

       Split out to make it easier to mock
    """
    with open(paths.HTTPD_IPA_PKI_PROXY_CONF, "r") as fd:
        lines = fd.readlines()

    return lines


@registry
class IPAProxySecretCheck(IPAPlugin):
    """
    Ensure that the proxy secrets match between tomcat and Apache

    Also report if tomcat has both secret and requiredSecret
    defined and whether all three secrets match.
    """
    @duration
    def check(self):
        if not self.ca.is_configured():
            logger.debug("CA is not configured, skipping IPAProxySecretCheck")
            return

        PROXY_SECRETS = 'proxy_secrets'

        # so many things can go wrong just keep one big global to
        # determine if we can eventually return SUCCESS
        failures = False

        server_xml = lxml.etree.parse(paths.PKI_TOMCAT_SERVER_XML)
        doc = server_xml.getroot()

        # no AJP connector means nothing to check
        connectors = doc.xpath('//Connector[@protocol="AJP/1.3"]')
        if len(connectors) == 0:
            yield Result(self, constants.CRITICAL,
                         key=PROXY_SECRETS,
                         server_xml=paths.PKI_TOMCAT_SERVER_XML,
                         msg='No AJP/1.3 Connectors defined in {server_xml}')
            return

        # IPA only deals with the first connect so that's all we'll check
        connector = connectors[0]

        ajp_secret = []
        if 'secret' in connector.attrib:
            ajp_secret.append(connector.attrib['secret'])

        if 'requiredSecret' in connector.attrib:
            ajp_secret.append(connector.attrib['requiredSecret'])

        if len(ajp_secret) > 1:
            if ajp_secret[0] != ajp_secret[1]:
                failures = True
                yield Result(
                    self, constants.WARNING,
                    key=PROXY_SECRETS,
                    server_xml=paths.PKI_TOMCAT_SERVER_XML,
                    msg='The AJP secrets in {server_xml} do not match'
                )
        # We could warn that both secret and requiredSecret are defined
        # but the presence of both with the same password doesn't
        # break anything so we will not warn for now.

        lines = read_ipa_pki_proxy()

        proxy_secrets = []
        PROXY_RE = r'\s+ProxyPassMatch ajp://localhost:8009 secret=(\w+)$'
        # Collect all the ipa-pki-proxy.conf secrets and ensure they all match
        for line in lines:
            m = re.match(PROXY_RE, line)
            if m:
                proxy_secrets.extend(m.groups(1))

        if not proxy_secrets:
            failures = True
            yield Result(
                self, constants.CRITICAL,
                key=PROXY_SECRETS,
                proxy_conf=paths.HTTPD_IPA_PKI_PROXY_CONF,
                msg='No ProxyPassMatch secrets found in {proxy_conf}'
            )
            return

        if len(set(proxy_secrets)) != 1:
            failures = True
            yield Result(
                self, constants.CRITICAL,
                key=PROXY_SECRETS,
                proxy_conf=paths.HTTPD_IPA_PKI_PROXY_CONF,
                msg='Not all ProxyPassMatch secrets match in {proxy_conf}'
            )

        for secret in proxy_secrets:
            if secret not in ajp_secret:
                failures = True
                yield Result(
                    self, constants.CRITICAL,
                    key=PROXY_SECRETS,
                    server_xml=paths.PKI_TOMCAT_SERVER_XML,
                    msg='A ProxyPassMatch secret not found in {server_xml}'
                )

        if not failures:
            yield Result(self, constants.SUCCESS,
                         key=PROXY_SECRETS)




# Copyright (C) 2019 FreeIPA Contributors see COPYING for license
#

from dns import rdatatype
from dns.exception import DNSException
import logging

from ipahealthcheck.ipa.plugin import IPAPlugin, registry
from ipahealthcheck.core.plugin import Result, duration
from ipahealthcheck.core import constants

from ipalib import api

try:
    from dns.resolver import resolve
except ImportError:
    from dns.resolver import query as resolve


logger = logging.getLogger()


def query_uri(uri):
    try:
        answers = resolve(uri, rdatatype.URI)
    except DNSException as e:
        logger.debug("DNS record not found: %s", e.__class__.__name__)
        answers = []
    return answers


@registry
class IPADNSSystemRecordsCheck(IPAPlugin):
    """
    Verify that the expected DNS service records are resolvable.

    IPA will already provide the values we need to validate with the
    IPASystemRecords class. We just need to pull that and do the
    equivalent DNS lookups.
    """
    requires = ('dirsrv',)

    def srv_to_name(self, srv, target):
        """Combine the SRV record and target into a unique name."""
        return srv + ":" + target

    def uri_to_name(self, uri, target):
        """Combine the SRV record and target into a unique name."""
        return uri + ":" + target

    @duration
    def check(self):
        # pylint: disable=import-outside-toplevel
        from ipapython.dnsutil import query_srv
        from ipaserver.dns_data_management import IPASystemRecords
        # pylint: enable=import-outside-toplevel

        system_records = IPASystemRecords(api)
        base_records = system_records.get_base_records()

        # collect the list of expected values
        txt_rec = dict()
        srv_rec = dict()
        uri_rec = dict()
        a_rec = list()
        aaaa_rec = list()

        for name, node in base_records.items():
            for rdataset in node:
                for rd in rdataset:
                    if rd.rdtype == rdatatype.SRV:
                        if name.ToASCII() in srv_rec:
                            srv_rec[name.ToASCII()].append(rd.target.to_text())
                        else:
                            srv_rec[name.ToASCII()] = [rd.target.to_text()]
                    elif rd.rdtype == rdatatype.TXT:
                        if name.ToASCII() in txt_rec:
                            txt_rec[name.ToASCII()].append(rd.to_text())
                        else:
                            txt_rec[name.ToASCII()] = [rd.to_text()]
                    elif rd.rdtype == rdatatype.A:
                        a_rec.append(rd.to_text())
                    elif rd.rdtype == rdatatype.AAAA:
                        aaaa_rec.append(rd.to_text())
                    elif rd.rdtype == rdatatype.URI:
                        if name.ToASCII() in uri_rec:
                            uri_rec[name.ToASCII()].append(
                                rd.target.decode('utf-8')
                            )
                        else:
                            uri_rec[name.ToASCII()] = [
                                rd.target.decode('utf-8')
                            ]
                    else:
                        logger.error("Unhandled rdtype %d", rd.rdtype)

        # For each SRV record that IPA thinks it should have, do a DNS
        # lookup of it and ensure that DNS has the same set of values
        # that IPA thinks it should.
        for srv, hosts in srv_rec.items():
            logger.debug("Search DNS for SRV record of %s", srv)
            try:
                answers = query_srv(srv)
            except DNSException as e:
                logger.debug("DNS record not found: %s", e.__class__.__name__)
                answers = []
            for answer in answers:
                logger.debug("DNS record found: %s", answer)
                try:
                    hosts.remove(answer.target.to_text())
                    yield Result(
                         self, constants.SUCCESS,
                         key=self.srv_to_name(srv, answer.target.to_text()))
                except ValueError:
                    yield Result(
                        self, constants.WARNING,
                        msg='Unexpected SRV entry in DNS',
                        key=self.srv_to_name(srv, answer.target.to_text()))
            for host in hosts:
                yield Result(
                    self, constants.WARNING,
                    msg='Expected SRV record missing',
                    key=self.srv_to_name(srv, host))

        for uri, hosts in uri_rec.items():
            logger.debug("Search DNS for URI record of %s", uri)
            answers = query_uri(uri)
            for answer in answers:
                logger.debug("DNS record found: %s", answer)
                try:
                    hosts.remove(answer.target.decode('utf-8'))
                    yield Result(
                         self, constants.SUCCESS,
                         key=self.uri_to_name(
                             uri, answer.target.decode('utf-8')
                         )
                    )
                except ValueError:
                    yield Result(
                        self, constants.WARNING,
                        msg='Unexpected URI entry in DNS',
                        key=self.uri_to_name(
                            uri, answer.target.decode('utf-8')
                        )
                    )
            for host in hosts:
                yield Result(
                    self, constants.WARNING,
                    msg='Expected URI record missing',
                    key=self.uri_to_name(uri, host)
                )

        for txt, realms in txt_rec.items():
            logger.debug("Search DNS for TXT record of %s", txt)
            try:
                answers = resolve(txt, rdatatype.TXT)
            except DNSException as e:
                logger.debug("DNS record not found: %s", e.__class__.__name__)
                answers = []

            for answer in answers:
                logger.debug("DNS record found: %s", answer)
                realm = answer.to_text()
                try:
                    realms.remove(realm)
                    yield Result(self, constants.SUCCESS,
                                 key=realm)
                except ValueError:
                    yield Result(self, constants.WARNING,
                                 key=realm,
                                 msg='expected realm missing')

        # Verify that all of the ipa-ca record IPs match those of
        # servers with a CA role. Report any missing or unexpected.
        qname = "ipa-ca." + api.env.domain + "."
        ipa_ca_records = []
        for dtype in (rdatatype.A, rdatatype.AAAA):
            logger.debug("Search DNS for %s records of %s", dtype.name, qname)
            try:
                answers = resolve(qname, dtype)
            except DNSException as e:
                logger.debug("DNS record not found: %s", e.__class__.__name__)
                answers = []
            for answer in answers:
                ipa_ca_records.append(answer.to_text())

        # Get the set of servers with the 'CA server' role
        ca_servers = {}
        for server in system_records.servers_data:
            host = system_records.servers_data.get(server)
            if 'CA server' in host.get('roles'):
                for dtype in (rdatatype.A, rdatatype.AAAA):
                    try:
                        a = resolve(server + '.', dtype)
                    except DNSException:
                        pass
                    else:
                        for answer in a:
                            if server in ca_servers:
                                ca_servers[server].append(answer.to_text())
                            else:
                                ca_servers[server] = [answer.to_text()]

        all_ca_ipaddr = []
        for server, ipaddrs in ca_servers.items():
            for ipaddr in ipaddrs:
                all_ca_ipaddr.append(ipaddr)

        # Loop through the ipa-ca records to determine if any are not
        # in the collection of all the reported CA server IPs.
        errors = 0
        for ipaddr in ipa_ca_records:
            if ipaddr not in all_ca_ipaddr:
                errors += 1
                yield Result(self, constants.WARNING,
                             key='ipa_ca_non_server_%s' % ipaddr,
                             ipaddr=ipaddr,
                             msg='Unexpected ipa-ca address {ipaddr}')

        # Remove any IP addresses we found for ipa-ca from the set of
        # IP addresses for all the IPA servers. Any remaining ones
        # are not in the ipa-ca A/AAAA record. We're only looking at
        # the DNS advertised servers so hidden ones should not be
        # here.
        for server, ipaddrs in ca_servers.items():
            for ipaddr in ipa_ca_records:
                if ipaddr in ipaddrs:
                    ipaddrs.remove(ipaddr)

        for server, ipaddrs in ca_servers.items():
            if ipaddrs:
                errors += 1
                yield Result(self, constants.WARNING,
                             key='ipa_ca_missing_%s' % server,
                             server=server,
                             ipaddr=', '.join(ipaddrs),
                             msg='expected ipa-ca to contain {ipaddr} for '
                                 '{server}')

        if errors == 0:
            yield Result(self, constants.SUCCESS, key='ipa_ca_check')



#
# Copyright (C) 2019 FreeIPA Contributors see COPYING for license
#

import uuid
from datetime import datetime
from functools import wraps

from ipahealthcheck.core.constants import getLevelName, getLevel


def duration(f):
    """Compute the duration of execution"""
    @wraps(f)
    def wrapper(*args, **kwds):
        start = datetime.utcnow()
        end = None
        for result in f(*args, **kwds):
            end = datetime.utcnow()
            dur = end - start
            result.duration = '%6.6f' % dur.total_seconds()
            yield result
        if end is None:
            # no results, yield None so a SUCCESS result will be created
            yield None
    return wrapper


class Registry:
    """
    A decorator that makes plugins available to the API

    Usage::

        register = Registry()

        @register()
        class some_plugin(...):
            ...
    """
    def __init__(self):
        self.plugins = []
        self.framework = None
        self.config = dict()
        self.options = None

    def initialize(self, framework, config, options=None):
        self.framework = framework
        self.config = config
        self.options = options

    def __call__(self, cls):
        if not callable(cls):
            raise TypeError('plugin must be callable; got %r' % cls)
        self.plugins.append(cls)
        return cls

    def get_plugins(self):
        for plugincls in self.plugins:
            yield plugincls(self)


class Plugin:
    """
    Base class for all plugins.

    registry defines where the plugin was registered, normally via
    a pkg_resource.

    requires is a tuple of strings that define pre-requisites for
    execution. Some output formats allow plugins that do not have
    these requirements met to skip them (JSON does NOT, all plugins
    are always executed and reported).

    Each Plugin should define a check() method that contains as
    simple a test as possible on the status a unique potential issue.

    A Plugin may return either Result for a single result or
    Results if multiple issues are discovered.

    It is strongly recommended to keep each Plugin as discrete as
    possible. This is not always possible or practical, for example
    to avoid hundreds of plugins that test nearly the same thing.

    Usage::

        register = Registry()

        @register()
        tmp_exists_check(Plugin)
            def check(self):
                if os.path.exists('/tmp'):
                    result = Result(self, SUCCESS)
                else:
                    result = Result(self, CRITICAL, path='/tmp',
                                    msg='Temporary directory is missing')

                return result

    """
    requires = ()

    def __init__(self, registry):
        self.registry = registry
        self.config = registry.config


class Result:
    """
    The result of a check.

    :param plugin: The plugin which generated the result.
    :param result: A result constant representing the level of error.
    :param source: If no plugin is passed then the name of the source
                   can be provided directly.
    :param check: If no plugin is passed then the name of the check
                   can be provided directly.
    :param kw: A dictionary of items providing insight in the error.

    Either both check and source need to be provided or plugin needs
    to be provided.

    kw is meant to provide some level of flexibility to check authors
    but the following is a set of pre-defined keys that may be present:

        key: some checks can have multiple tests. This
             provides for uniqueuess.
        msg: A message that can take other keywords as input
        exception: used when a check raises an exception
    """
    def __init__(self, plugin, result, source=None, check=None,
                 start=None, duration=None, when=None, **kw):
        self.result = result
        self.kw = kw
        self.when = when or generalized_time(datetime.utcnow())
        self.duration = duration
        self.uuid = str(uuid.uuid4())
        if None not in (check, source):
            self.check = check
            self.source = source
        else:
            if plugin is None:
                raise TypeError('source and check or plugin must be provided')
            self.check = plugin.__class__.__name__
            self.source = plugin.__class__.__module__
        if start is not None:
            dur = datetime.utcnow() - start
            self.duration = '%6.6f' % dur.total_seconds()

        assert getLevelName(result) is not None

    def __repr__(self):
        return "%s.%s(%s): %s" % (self.source, self.check, self.kw,
                                  self.result)


class Results:
    """
    A list-like collection of Result values.

    Provides a very limited subset of list operations. Is intended for
    internal-use only and not by check functions.

    Usage::

        results = Results()

        result = Result(plugin, SUCCESS, **kw)
        results.add(result)
    """
    def __init__(self):
        self.results = []

    def __len__(self):
        return len(self.results)

    def add(self, result):
        assert isinstance(result, Result)
        self.results.append(result)

    def extend(self, results):
        assert isinstance(results, Results)
        self.results.extend(results.results)

    def output(self):
        for result in self.results:
            yield dict(source=result.source,
                       check=result.check,
                       result=getLevelName(result.result),
                       uuid=result.uuid,
                       when=result.when,
                       duration=result.duration,
                       kw=result.kw)


def json_to_results(data):
    """
    Convert JSON data into a Results object.

    :param data: valid JSON input
    :returns: a Results object representing the JSON input
    """

    results = Results()

    for line in data:
        result = getLevel(line.pop('result'))
        source = line.pop('source')
        check = line.pop('check')
        duration = line.pop('duration')
        when = line.pop('when')
        kw = line.pop('kw')
        result = Result(None, result, source, check, duration=duration,
                        when=when, **kw)
        results.add(result)

    return results


def generalized_time(intime):
    """Convert a datetime.datetime object to LDAP generalized time format

       :param intime: a datetime.datetime object
    """
    assert isinstance(intime, datetime)
    return intime.strftime('%Y%m%d%H%M%SZ')



#
# Copyright (C) 2019 FreeIPA Contributors see COPYING for license
#

from util import capture_results, CAInstance, KRAInstance
from base import BaseTest
from ipahealthcheck.core import config, constants
from ipahealthcheck.dogtag.plugin import registry
from ipahealthcheck.dogtag.ca import DogtagCertsConfigCheck
from unittest.mock import Mock, patch


class mock_Cert:
    """Fake up a certificate.

      The contents are the NSS nickname of the certificate.
    """
    def __init__(self, text):
        self.text = text

    def public_bytes(self, encoding):
        return self.text.encode('utf-8')


class mock_CertDB:
    def __init__(self, trust):
        """A dict of nickname + NSSdb trust flags"""
        self.trust = trust

    def list_certs(self):
        return [(nickname, self.trust[nickname]) for nickname in self.trust]

    def get_cert_from_db(self, nickname):
        """Return the nickname. This will match the value of get_directive"""
        return mock_Cert(nickname)


class TestCACerts(BaseTest):
    patches = {
        'ipaserver.install.cainstance.CAInstance':
        Mock(return_value=CAInstance()),
        'ipaserver.install.krainstance.KRAInstance':
        Mock(return_value=KRAInstance()),
    }

    @patch('ipahealthcheck.dogtag.ca.get_directive')
    @patch('ipaserver.install.certs.CertDB')
    def test_ca_certs_ok(self, mock_certdb, mock_directive):
        """Test what should be the standard case"""
        trust = {
            'ocspSigningCert cert-pki-ca': 'u,u,u',
            'subsystemCert cert-pki-ca': 'u,u,u',
            'auditSigningCert cert-pki-ca': 'u,u,Pu',
            'Server-Cert cert-pki-ca': 'u,u,u',
            'caSigningCert cert-pki-ca': 'CT,C,C',
            'transportCert cert-pki-kra': 'u,u,u',
        }
        mock_certdb.return_value = mock_CertDB(trust)
        mock_directive.side_effect = [name for name, nsstrust in trust.items()]

        framework = object()
        registry.initialize(framework, config.Config())
        f = DogtagCertsConfigCheck(registry)

        self.results = capture_results(f)

        assert len(self.results) == 6

        for result in self.results.results:
            assert result.result == constants.SUCCESS
            assert result.source == 'ipahealthcheck.dogtag.ca'
            assert result.check == 'DogtagCertsConfigCheck'

    @patch('ipahealthcheck.dogtag.ca.get_directive')
    @patch('ipaserver.install.certs.CertDB')
    def test_cert_missing_from_file(self, mock_certdb, mock_directive):
        """Test a missing certificate.

           Note that if it is missing from the database then this check
           will not catch the error but it will be caught elsewhere.
        """
        trust = {
            'ocspSigningCert cert-pki-ca': 'u,u,u',
            'subsystemCert cert-pki-ca': 'u,u,u',
            'auditSigningCert cert-pki-ca': 'u,u,Pu',
            'Server-Cert cert-pki-ca': 'u,u,u',
            'caSigningCert cert-pki-ca': 'CT,,',
            'transportCert cert-pki-kra': 'u,u,u',
        }

        # The 3rd cert won't match the results
        nicknames = [name for name, nsstrust in trust.items()]
        location = nicknames.index('auditSigningCert cert-pki-ca')
        nicknames[location] = 'NOT auditSigningCert cert-pki-ca'

        mock_certdb.return_value = mock_CertDB(trust)
        mock_directive.side_effect = nicknames

        framework = object()
        registry.initialize(framework, config.Config)
        f = DogtagCertsConfigCheck(registry)

        self.results = capture_results(f)

        num = len(self.results.results)
        for r in range(0, num):
            if r == 2:  # skip the one that should be bad
                continue
            result = self.results.results[r]
            assert result.result == constants.SUCCESS
            assert result.source == 'ipahealthcheck.dogtag.ca'
            assert result.check == 'DogtagCertsConfigCheck'

        result = self.results.results[2]

        assert result.result == constants.ERROR
        assert result.source == 'ipahealthcheck.dogtag.ca'
        assert result.check == 'DogtagCertsConfigCheck'
        assert result.kw.get('key') == 'auditSigningCert cert-pki-ca'

        assert len(self.results) == 6

    @patch('ipaserver.install.cainstance.CAInstance')
    def test_cacert_caless(self, mock_cainstance):
        """Nothing to check if the master is CALess"""

        mock_cainstance.return_value = CAInstance(False)

        framework = object()
        registry.initialize(framework, config)
        f = DogtagCertsConfigCheck(registry)

        self.results = capture_results(f)

        assert len(self.results) == 0



#
# Copyright (C) 2019 FreeIPA Contributors see COPYING for license
#

from util import capture_results
from base import BaseTest

from ipahealthcheck.core import constants, config
from ipahealthcheck.ipa.plugin import registry
from ipahealthcheck.ipa.certs import IPACertTracking
from unittest.mock import Mock
from mock_certmonger import create_mock_dbus, _certmonger
from mock_certmonger import get_expected_requests, set_requests


class TestTracking(BaseTest):
    patches = {
        'ipahealthcheck.ipa.certs.get_expected_requests':
        Mock(return_value=get_expected_requests()),
        'ipalib.install.certmonger._cm_dbus_object':
        Mock(side_effect=create_mock_dbus),
        'ipalib.install.certmonger._certmonger':
        Mock(return_value=_certmonger())
    }

    def test_known_cert_tracking(self):
        set_requests()

        framework = object()
        registry.initialize(framework, config.Config)
        f = IPACertTracking(registry)

        self.results = capture_results(f)

        assert len(self.results) == 2

    def test_missing_cert_tracking(self):
        # remove one of the requests to force it to be missing
        set_requests(remove=0)

        framework = object()
        registry.initialize(framework, config.Config)
        f = IPACertTracking(registry)

        self.results = capture_results(f)

        assert len(self.results) == 2

        result = self.results.results[0]
        assert result.result == constants.ERROR
        assert result.source == 'ipahealthcheck.ipa.certs'
        assert result.check == 'IPACertTracking'
        assert result.kw.get('key') == \
            "cert-file=/var/lib/ipa/ra-agent.pem, " \
            "key-file=/var/lib/ipa/ra-agent.key, " \
            "ca-name=dogtag-ipa-ca-renew-agent, " \
            "template_profile=caSubsystemCert, " \
            "cert-storage=FILE, "\
            "cert-presave-command=" \
            "/usr/libexec/ipa/certmonger/renew_ra_cert_pre, " \
            "cert-postsave-command=" \
            "/usr/libexec/ipa/certmonger/renew_ra_cert"

    def test_unknown_cert_tracking(self):
        # Add a custom, unknown request
        unknown = {
            'nickname': '7777',
            'cert-file': '/tmp/test.crt',
            'key-file': '/tmp/test.key',
            'ca-name': 'IPA',
        }
        set_requests(add=unknown)

        framework = object()
        registry.initialize(framework, config.Config)
        f = IPACertTracking(registry)

        self.results = capture_results(f)

        assert len(self.results) == 3

        result = self.results.results[2]
        assert result.result == constants.WARNING
        assert result.source == 'ipahealthcheck.ipa.certs'
        assert result.check == 'IPACertTracking'
        assert result.kw.get('key') == '7777'



#
# Copyright (C) 2019 FreeIPA Contributors see COPYING for license
#

from ipaclustercheck.ipa.plugin import ClusterPlugin, registry, find_checks
from ipahealthcheck.core.plugin import Result, duration
from ipahealthcheck.core import constants


@registry
class ClusterCRLManagerCheck(ClusterPlugin):

    @duration
    def check(self):
        data = self.registry.json
        crlmanagers = []

        for fqdn in data.keys():
             output = find_checks(data[fqdn], 'ipahealthcheck.ipa.roles',
                                 'IPACRLManagerCheck')
             enabled = output[0].get('kw').get('crlgen_enabled')
             if enabled:
                 crlmanagers.append(fqdn)
        if len(crlmanagers) == 0:
            yield Result(self, constants.ERROR,
                         name='crlmanager',
                         error='No CRL Manager defined')
        elif len(crlmanagers) == 1:
            yield Result(self, constants.SUCCESS,
                         name='crlmanager',
                         value=crlmanagers[0])
        else:
            yield Result(self, constants.ERROR,
                         name='crlmanager',
                         value=','.join(crlmanagers),
                         error='Multiple CRL Managers defined')



#
# Copyright (C) 2020 FreeIPA Contributors see COPYING for license
#

from ipahealthcheck.ds.plugin import DSPlugin, registry
from lib389.plugins import ReferentialIntegrityPlugin


@registry
class RIPluginCheck(DSPlugin):
    """
    Check that the RI plugin configuration is valid and properly indexed
    """
    check_class = ReferentialIntegrityPlugin



#
# Copyright (C) 2020 FreeIPA Contributors see COPYING for license
#

import logging
import re
from urllib.parse import urlparse

from ipahealthcheck.ds.plugin import DSPlugin, registry
from ipahealthcheck.core.plugin import Result
from ipahealthcheck.core.plugin import duration
from ipahealthcheck.core import constants

from ipalib import api, errors
from ipapython.dn import DN

logger = logging.getLogger()


@registry
class RUVCheck(DSPlugin):
    """
    Provide the main and dogtag RUV.

    Local analysis is not possible since it requires collecting the
    RUV from all masters and healthcheck is limited to only talking
    to itself.
    """
    requires = ('dirsrv',)

    def get_ruv(self, dn):
        """Identify the RUV for a suffix on this master"""
        try:
            entry = self.conn.get_entry(dn)
        except Exception:
            return None
        else:
            return entry.single_value.get('nsDS5ReplicaID')

    @duration
    def check(self):
        ruv = self.get_ruv(DN(('cn', 'replica'), ('cn', api.env.basedn),
                           ('cn', 'mapping tree'), ('cn', 'config')))
        csruv = self.get_ruv(DN(('cn', 'replica'), ('cn', 'o=ipaca'),
                             ('cn', 'mapping tree'), ('cn', 'config')))

        if ruv is not None:
            yield Result(self, constants.SUCCESS,
                         key=str(api.env.basedn),
                         ruv=ruv)
        if csruv is not None:
            yield Result(self, constants.SUCCESS,
                         key='o=ipaca',
                         ruv=csruv)


@registry
class KnownRUVCheck(DSPlugin):
    """Return all known RUVs. This can be used to identify "dangling"
       RUVs, or left-overs from previous replication agreements.
    """
    requires = ('dirsrv',)

    def get_all_ruvs(self, suffix):
        """Get all known RUVs on this master

           Return the RUV entries as a list of tuples: (hostname, rid)
        """
        search_filter = '(&(nsuniqueid=ffffffff-ffffffff-ffffffff-ffffffff)' \
                        '(objectclass=nstombstone))'
        try:
            entries = self.conn.get_entries(
                suffix, self.conn.SCOPE_SUBTREE, search_filter,
                ['nsds50ruv'])
        except errors.NotFound:
            logger.debug("No RUV records found.")
            return []
            # raise NoRUVsFound("No RUV records found.")

        servers = []
        for e in entries:
            for ruv in e['nsds50ruv']:
                if ruv.startswith('{replicageneration'):
                    continue
                data = re.match(
                    r'\{replica (\d+) (ldap://.*:\d+)\}(\s+\w+\s+\w*){0,1}',
                    ruv
                )
                if data:
                    rid = data.group(1)
                    (
                        _scheme, netloc, _path, _params, _query, _fragment
                    ) = urlparse(data.group(2))
                    servers.append((re.sub(r':\d+', '', netloc), rid))
                else:
                    logger.debug("Unable to decode RUV: %s", ruv)

        return servers

    @duration
    def check(self):

        ruvs = self.get_all_ruvs(api.env.basedn)
        csruvs = self.get_all_ruvs(DN('o=ipaca'))

        if ruvs:
            yield Result(self, constants.SUCCESS,
                         key='ruvs_' + str(api.env.basedn),
                         suffix=str(api.env.basedn),
                         ruvs=ruvs)

        if csruvs:
            yield Result(self, constants.SUCCESS,
                         key='ruvs_o=ipaca',
                         suffix='o=ipaca',
                         ruvs=csruvs)



#
# Copyright (C) 2020 FreeIPA Contributors see COPYING for license
#

from ipahealthcheck.ds.plugin import DSPlugin, registry
from lib389.dseldif import DSEldif


@registry
class DSECheck(DSPlugin):
    """
    Check the dse.ldif/cn=config for obvious issues
    """
    check_class = DSEldif



#
# Copyright (C) 2019 FreeIPA Contributors see COPYING for license
#

import logging

from ipaclustercheck.ipa.plugin import (
    ClusterPlugin,
    registry,
    find_checks,
    get_masters
)
from ipahealthcheck.core.plugin import Result, duration
from ipahealthcheck.core import constants
from ipalib import api
from ipapython.dn import DN


logger = logging.getLogger()


@registry
class ClusterRUVCheck(ClusterPlugin):

    # TODO: confirm that all masters are represented, otherwise the
    #       trustworthiness of dangling RUV is mixed.
    #
    #       gah, need to provide full list of all masters in a check.

    @duration
    def check(self):
        data = self.registry.json

        # Start with the list of masters from the file(s) collected
        # and find a MetaCheck with a full list of masters. For
        # backwards compatibility.
        try:
            masters = get_masters(data)
        except ValueError as e:
            yield Result(self, constants.ERROR,
                         name='dangling_ruv',
                         error=str(e))
            return

        if len(data.keys()) < len(masters):
            yield Result(self, constants.ERROR,
                         name='dangling_ruv',
                         error='Unable to determine list of RUVs, missing '
                               'some masters: %s' %
                               ''.join(set(masters) - set(data.keys())))
            return

        # collect the full set of known RUVs for each master
        info = {}
        for master in masters:
            info[master] = {
                'ca': False,           # does the host have ca configured?
                'ruvs': set(),         # ruvs on the host
                'csruvs': set(),       # csruvs on the host
                'clean_ruv': set(),    # ruvs to be cleaned from the host
                'clean_csruv': set()   # csruvs to be cleaned from the host
                }

        for fqdn in data.keys():
            outputs = find_checks(data[fqdn], 'ipahealthcheck.ds.ruv',
                                  'KnownRUVCheck')
            for output in outputs:
                if not 'suffix' in output.get('kw'):
                    continue
                basedn = DN(output.get('kw').get('suffix'))

                ruvset = set()
                ruvtmp = output.get('kw').get('ruvs')
                for ruv in ruvtmp:
                    ruvset.add(tuple(ruv))

                if basedn == DN('o=ipaca'):
                    info[fqdn]['ca'] = True
                    info[fqdn]['csruvs'] = ruvset
                elif basedn == api.env.basedn:
                    info[fqdn]['ruvs'] = ruvset
                else:
                    yield Result(self, constants.WARNING,
                                 name='dangling_ruv',
                                 error='Unknown suffix found %s expected %s'
                                       % (basedn, api.env.basedn))

        # Collect the nsDS5ReplicaID for each master
        ruvs = set()
        csruvs = set()
        for fqdn in data.keys():
            outputs = find_checks(data[fqdn], 'ipahealthcheck.ds.ruv',
                                  'RUVCheck')
            for output in outputs:
                if not 'key' in output.get('kw'):
                    continue
                basedn = DN(output.get('kw').get('key'))
                ruv = (fqdn, (output.get('kw').get('ruv')))
                if basedn == DN('o=ipaca'):
                    csruvs.add(ruv)
                elif basedn == api.env.basedn:
                    ruvs.add(ruv)
                else:
                    yield Result(self, constants.WARNING,
                                 name='dangling_ruv',
                                 error='Unknown suffix found %s expected %s'
                                       % (basedn, api.env.basedn))

        dangles = False
        # get the dangling RUVs
        for master_info in info.values():
            for ruv in master_info['ruvs']:
                if ruv not in ruvs:
                    master_info['clean_ruv'].add(ruv)
                    dangles = True

            # if ca is not configured, there will be no csruvs in master_info
            for csruv in master_info['csruvs']:
                if csruv not in csruvs:
                    master_info['clean_csruv'].add(csruv)
                    dangles = True

        clean_csruvs = set()
        clean_ruvs = set()
        if dangles:
            for _unused, master_info in info.items():
                for ruv in master_info['clean_ruv']:
                    logger.debug(
                        "Dangling RUV id: %s, hostname: %s", ruv[1], ruv[0]
                    )
                    clean_ruvs.add(ruv[1])
                for csruv in master_info['clean_csruv']:
                    logger.debug(
                        "Dangling CS RUV id: %s, hostname: %s",
                        csruv[1],
                        csruv[0]
                    )
                    clean_csruvs.add(csruv[1])

        if clean_ruvs:
            yield Result(self, constants.ERROR,
                         name='dangling_ruv',
                         value=', '.join(clean_ruvs))
        else:
            yield Result(self, constants.SUCCESS,
                         name='dangling_ruv',
                         value='No dangling RUVs found')
        if clean_csruvs:
            yield Result(self, constants.ERROR,
                         name='dangling_csruv',
                         value=', '.join(clean_csruvs))
        else:
            yield Result(self, constants.SUCCESS,
                         name='dangling_csruv',
                         value='No dangling CS RUVs found')



#
# Copyright (C) 2019 FreeIPA Contributors see COPYING for license
#

from ipahealthcheck.core.plugin import Plugin, Registry


class MetaPlugin(Plugin):
    pass


class MetaRegistry(Registry):
    pass


registry = MetaRegistry()



#
# Copyright (C) 2019 FreeIPA Contributors see COPYING for license
#

from ipahealthcheck.core.plugin import Plugin, Registry


class SystemPlugin(Plugin):
    pass


class SystemRegistry(Registry):
    def initialize(self, framework, config, options=None):
        pass


registry = SystemRegistry()



#
# Copyright (C) 2019 FreeIPA Contributors see COPYING for license
#

import argparse
import json
import logging
import pkg_resources
import signal
import sys
import traceback
import warnings

from datetime import datetime

from ipahealthcheck.core.config import read_config
from ipahealthcheck.core.exceptions import TimeoutError
from ipahealthcheck.core.plugin import Result, Results, json_to_results
from ipahealthcheck.core.output import output_registry
from ipahealthcheck.core import constants
from ipahealthcheck.core.service import ServiceCheck

logging.basicConfig(format='%(message)s')
logger = logging.getLogger()


def find_registries(entry_points):
    # Loading the resources may reset the log level, save it.
    log_level = logger.level
    registries = {}
    for entry_point in entry_points:
        registries.update({
            ep.name: ep.resolve()
            for ep in pkg_resources.iter_entry_points(entry_point)
        })
    logger.setLevel(log_level)
    return registries


def find_plugins(name, registry):
    for ep in pkg_resources.iter_entry_points(name):
        # load module
        ep.load()
    return registry.get_plugins()


def run_plugin(plugin, available=(), timeout=constants.DEFAULT_TIMEOUT):
    def signal_handler(signum, frame):
        raise TimeoutError('Request timed out')

    # manually calculate duration when we create results of our own
    start = datetime.utcnow()
    signal.signal(signal.SIGALRM, signal_handler)
    signal.alarm(timeout)
    try:
        for result in plugin.check():
            if result is None:
                # Treat no result as success, fudge start time
                result = Result(plugin, constants.SUCCESS, start=start)
            yield result
    except TimeoutError as e:
        yield Result(plugin, constants.ERROR, exception=str(e),
                     start=start)
    except Exception as e:
        logger.debug('Exception raised: %s', e)
        logger.debug(traceback.format_exc())
        yield Result(plugin, constants.CRITICAL, exception=str(e),
                     traceback=traceback.format_exc(),
                     start=start)
    finally:
        signal.alarm(0)


def source_or_check_matches(plugin, source, check):
    """Determine whether a given a plugin matches if a source
       and optional check are provided.
    """
    if (
        source is not None and
        not _is_prefix_of_source(source, plugin.__module__)
    ):
        return False

    if check and plugin.__class__.__name__ != check:
        return False

    return True


def exclude_source_or_check(source, check, config):
    """Return True if a source or check should be excluded, otherwise False"""
    exclude_source = []
    exclude_check = []

    if 'excludes_source' in config:
        exclude_source = config.excludes_source
    if 'excludes_check' in config:
        exclude_check = config.excludes_check

    for exclude in exclude_source:
        if _is_prefix_of_source(exclude, source):
            return True

    for exclude in exclude_check:
        if exclude == check:
            return True

    return False


def run_service_plugins(plugins, source, check):
    """Execute plugins with the base class of ServiceCheck

       This is a specialized check to use systemd to determine
       if a service is running or not.
    """
    results = Results()
    available = []

    for plugin in plugins:
        if not isinstance(plugin, ServiceCheck):
            continue

        # Try to save some time to not check dependent services if the
        # parent is down.
        if not set(plugin.requires).issubset(available):
            # A required service is not available. Either it hasn't been
            # checked yet or it isn't running. If not running break.
            running = True
            for result in results.results:
                if result.check in plugin.requires:
                    # if not in available but in results the service failed
                    running = False
                    break
            if not running:
                logger.debug(
                    'Skipping %s:%s because %s service(s) not running',
                    plugin.__class__.__module__,
                    plugin.__class__.__name__,
                    ', '.join(set(plugin.requires) - set(available))
                )
                continue

        logger.debug('Calling check %s', plugin)
        for result in plugin.check():
            # always run the service checks so dependencies work
            if result is not None and result.result == constants.SUCCESS:
                available.append(plugin.service.service_name)
            if not source_or_check_matches(plugin, source, check):
                continue
            if result is not None:
                results.add(result)

    return results, set(available)


def run_plugins(plugins, available, source, check,
                config, timeout=constants.DEFAULT_TIMEOUT):
    """Execute plugins without the base class of ServiceCheck

       These are the remaining, non-service checking checks
       that do validation for various parts of a system.
    """
    results = Results()

    for plugin in plugins:
        if isinstance(plugin, ServiceCheck):
            continue

        if exclude_source_or_check(
            plugin.__module__, plugin.__class__.__name__, config
        ):
            logger.debug("Excluding %s::%s per config",
                         plugin.__module__, plugin.__class__.__name__)
            continue
        if not source_or_check_matches(plugin, source, check):
            continue

        logger.debug("Calling check %s", plugin)
        if not set(plugin.requires).issubset(available):
            logger.debug('Skipping %s:%s because %s service(s) not running',
                         plugin.__class__.__module__,
                         plugin.__class__.__name__,
                         ', '.join(set(plugin.requires) - available))
            # Not providing a Result in this case because if a required
            # service isn't available then this could generate a lot of
            # false positives.
        else:
            for result in run_plugin(plugin, available, timeout):
                results.add(result)

    return results


def list_sources(plugins):
    """Print list of all sources and checks"""
    source = None
    for plugin in plugins:
        if source != plugin.__class__.__module__:
            print(plugin.__class__.__module__)
            source = plugin.__class__.__module__
        print("  ", plugin.__class__.__name__)

    return 0


def add_default_options(parser, output_registry, default_output):
    output_names = [plugin.__name__.lower() for
                    plugin in output_registry.plugins]
    parser.add_argument('--config', dest='config',
                        default=None, help='Config file to load')
    parser.add_argument('--verbose', dest='verbose', action='store_true',
                        default=False, help='Run in verbose mode')
    parser.add_argument('--debug', dest='debug', action='store_true',
                        default=False, help='Include debug output')
    parser.add_argument('--list-sources', dest='list_sources',
                        action='store_true', default=False,
                        help='List all available sources')
    parser.add_argument('--source', dest='source',
                        default=None,
                        help='Source of checks, e.g. foo.bar.baz')
    parser.add_argument('--check', dest='check',
                        default=None,
                        help='Check to execute, e.g. BazCheck')
    parser.add_argument('--output-type', dest='output_type',
                        choices=output_names,
                        default=default_output, help='Output method')
    parser.add_argument('--output-file', dest='output_file', default=None,
                        help='File to store output')
    parser.add_argument('--version', dest='version', action='store_true',
                        help='Report the version number and exit')


def add_output_options(parser, output_registry):
    for plugin in output_registry.plugins:
        onelinedoc = plugin.__doc__.split('\n\n', 1)[0].strip()
        group = parser.add_argument_group(plugin.__name__.lower(),
                                          onelinedoc)
        for option in plugin.options:
            group.add_argument(option[0], **option[1])


def parse_options(parser):
    options = parser.parse_args()

    # Validation
    if options.check and not options.source:
        raise ValueError("--source is required when --check is used")

    return options


def limit_results(results, source, check):
    """Return ony those results which match source and/or check"""
    new_results = Results()
    for result in results.results:
        if check is None:
            # treat 'source' as prefix
            if _is_prefix_of_source(source, result.source):
                new_results.add(result)
        else:
            # when 'check' is given, match source fully
            if result.source == source and result.check == check:
                new_results.add(result)
    return new_results


def exclude_keys(config, results):
    """Generate a new result, excluding unwanted keys"""
    new_results = Results()

    for result in results.results:
        if (
            'excludes_key' in config and
            result.kw.get("key") in config.excludes_key
        ):
            logger.debug("Excluding %s::%s::%s per config",
                         result.source, result.check, result.kw.get('key'))
        else:
            new_results.add(result)

    return new_results


def _is_prefix_of_source(prefix, source):
    prefix_parts = prefix.split('.')
    source_parts = source.split('.')
    return source_parts[:len(prefix_parts)] == prefix_parts


class RunChecks:
    def __init__(self, entry_points, configfile,
                 output_registry=output_registry,
                 default_output='json'):
        """Initialize class variables

          entry_points: A list of entry points to find plugins
          configfile: full path to the config file
          output_registry: registry containing the set of output
                           plugins to register.
          default_output: default output class
        """
        self.entry_points = entry_points
        self.configfile = configfile
        self.output_registry = output_registry
        self.default_output = default_output
        self.parser = argparse.ArgumentParser()
        self.options = None

    def pre_check(self):
        return None

    def add_options(self):
        """Add custom options for this check program"""

    def validate_options(self):
        """Validate options other than source and check"""
        return None

    def run_healthcheck(self):
        framework = object()
        plugins = []
        output = None

        logger.setLevel(logging.WARNING)

        add_default_options(self.parser, self.output_registry,
                            self.default_output)
        add_output_options(self.parser, self.output_registry)
        self.add_options()
        try:
            options = parse_options(self.parser)
        except ValueError as e:
            sys.exit(str(e))

        if options.version:
            for registry in self.entry_points:
                name = registry.split('.')[0]
                try:
                    version = pkg_resources.get_distribution(name).version
                except pkg_resources.DistributionNotFound:
                    continue
                print('%s: %s' % (name, version))
            return 0

        # pylint: disable=assignment-from-none
        rval = self.validate_options()
        # pylint: enable=assignment-from-none
        if rval is not None:
            return rval

        if options.config is not None:
            config = read_config(options.config)
        else:
            config = read_config(self.configfile)
        if config is None:
            return 1

        # Unify config and options. One of these variables will be
        # eventually deprecated in the future. This way all cli
        # options can be set in config instead.
        config.merge(vars(options))
        self.options = config
        options = config

        if options.verbose:
            logger.setLevel(logging.INFO)

        if options.debug:
            logger.setLevel(logging.DEBUG)

        # pylint: disable=assignment-from-none
        rval = self.pre_check()
        # pylint: enable=assignment-from-none
        if rval is not None:
            return rval

        # The pki checks are noisy if a CA is not configured so we
        # want to suppress that for IPA.
        #
        # There are 3 possible states:
        # 1. IPA is configured with a CA
        # 2. IPA is configured without a CA
        # 3. IPA is not configured
        #
        # If we have IPA configured without a CA then we want to skip
        # the pkihealthcheck plugins
        #
        # The IPA registry will set ca_configured in its registry to True
        # or False. We will skip the pkihealthcheck plugins only if
        # ca_configured is False which means that it was set by IPA. So
        # we initialize ca_configured to None so that the pki checks
        # will always be executed with pki-healthcheck.
        ca_configured = None
        for name, registry in find_registries(self.entry_points).items():
            try:
                registry.initialize(framework, config, options)
            except Exception as e:
                warnings.warn("Trying deprecated initialization API: %s" % e,
                              DeprecationWarning)
                try:
                    registry.initialize(framework, config)
                except Exception as e:
                    logger.error("Unable to initialize %s: %s", name, e)
                    continue
            if hasattr(registry, 'ca_configured'):
                ca_configured = registry.ca_configured
        for name, registry in find_registries(self.entry_points).items():
            if 'pkihealthcheck' in name and ca_configured is False:
                logger.debug('IPA CA is not configured, skipping %s', name)
                continue
            for plugin in find_plugins(name, registry):
                plugins.append(plugin)

        for out in self.output_registry.plugins:
            if out.__name__.lower() == options.output_type:
                output = out(options)
                break
        if output is None:
            print(f"Unknown output-type '{options.output_type}'")
            return 1

        if options.list_sources:
            return list_sources(plugins)

        if 'infile' in options and options.infile:
            try:
                with open(options.infile, 'r') as f:
                    raw_data = f.read()

                json_data = json.loads(raw_data)
                results = json_to_results(json_data)
                available = ()
            except Exception as e:
                print("Unable to import '%s': %s" % (options.infile, e))
                return 1
            if options.source:
                results = limit_results(results, options.source, options.check)
        else:
            results, available = run_service_plugins(plugins,
                                                     options.source,
                                                     options.check)
            results.extend(run_plugins(plugins, available,
                                       options.source, options.check, config,
                                       config.timeout))

        if options.source and len(results.results) == 0:
            for plugin in plugins:
                if not source_or_check_matches(plugin, options.source,
                                               options.check):
                    continue

                if not set(plugin.requires).issubset(available):
                    print("Source '%s' is missing one or more requirements "
                          "'%s'" %
                          (options.source, ', '.join(plugin.requires)))
                    return 1

            if options.check:
                print("Check '%s' not found in Source '%s'" %
                      (options.check, options.source))
            else:
                print("Source '%s' not found" % options.source)
            return 1

        results = exclude_keys(config, results)

        try:
            output.render(results)
        except Exception as e:
            logger.error('Output raised %s: %s', e.__class__.__name__, e)

        return_value = 0
        for result in results.results:
            if result.result != constants.SUCCESS:
                return_value = 1
                break

        return return_value



#
# Copyright (C) 2019 FreeIPA Contributors see COPYING for license
#
from __future__ import division

from datetime import datetime, timezone, timedelta
import itertools
from inspect import signature
import logging
import os
import socket
import tempfile

from ipahealthcheck.ipa.plugin import IPAPlugin, registry
from ipahealthcheck.core.plugin import Result, generalized_time
from ipahealthcheck.core.plugin import duration
from ipahealthcheck.core import constants

from ipalib import api
from ipalib import errors
from ipalib import x509
from ipalib.install import certmonger
from ipalib.constants import RENEWAL_CA_NAME, IPA_CA_RECORD
from ipaplatform.paths import paths
from ipaserver.install import certs
from ipaserver.install import dsinstance
from ipaserver.install import krainstance
from ipaserver.install import krbinstance
from ipaserver.plugins import ldap2
from ipapython import certdb
from ipapython import ipautil
from ipapython.dn import DN
from ipapython.ipaldap import realm_to_serverid

logger = logging.getLogger()
DAY = 60 * 60 * 24


def is_ipa_issued_cert(myapi, cert):
    """Thin wrapper around certs.is_ipa_issued to test for LDAP"""
    if not myapi.Backend.ldap2.isconnected():
        return None

    return certs.is_ipa_issued_cert(myapi, cert)


def get_expected_requests(ca, ds, serverid):
    """Provide the expected certmonger tracking request data

       This list is based in part on certificate_renewal_update() in
       ipaserver/install/server/upgrade.py and various
       start_tracking_certificates() methods in *instance.py.

       The list is filtered depending on whether a CA is running
       and the certificates have been issued by IPA.

      :param ca: the CAInstance
      :param ds: the DSInstance
      :param serverid: the DS serverid name
    """
    template = paths.CERTMONGER_COMMAND_TEMPLATE

    if api.Command.ca_is_enabled()['result']:
        requests = [
            {
                'cert-file': paths.RA_AGENT_PEM,
                'key-file': paths.RA_AGENT_KEY,
                'ca-name': RENEWAL_CA_NAME,
                'cert-presave-command': template % 'renew_ra_cert_pre',
                'cert-postsave-command': template % 'renew_ra_cert',
            },
        ]
    else:
        requests = []

    if ca.is_configured():
        dogtag_reqs = ca.tracking_reqs.items()
        kra = krainstance.KRAInstance(api.env.realm)
        if kra.is_installed():
            dogtag_reqs = itertools.chain(dogtag_reqs,
                                          kra.tracking_reqs.items())
        for nick, profile in dogtag_reqs:
            if profile in ('caSignedLogCert', 'caOCSPCert',
                           'caSubsystemCert', 'caCACert',
                           'caAuditSigningCert', 'caTransportCert',
                           'caStorageCert'):
                token = ca.token_name
            else:
                token = None

            req = {
                'cert-database': paths.PKI_TOMCAT_ALIAS_DIR,
                'cert-nickname': nick,
                'ca-name': RENEWAL_CA_NAME,
                'cert-presave-command': template % 'stop_pkicad',
                'cert-postsave-command':
                    (template % 'renew_ca_cert "{}"'.format(nick)),
                'template-profile': profile,
            }
            if token and token != 'internal':
                req['key-token'] = token
            requests.append(req)
    else:
        logger.debug('CA is not configured, skipping CA tracking')

    cert = x509.load_certificate_from_file(paths.HTTPD_CERT_FILE)
    issued = is_ipa_issued_cert(api, cert)
    if issued is None:
        logger.debug('Unable to determine if \'%s\' was issued by IPA '
                     'because no LDAP connection, assuming yes.')
    if issued or issued is None:
        requests.append(
            {
                'cert-file': paths.HTTPD_CERT_FILE,
                'key-file': paths.HTTPD_KEY_FILE,
                'ca-name': 'IPA',
                'cert-postsave-command': template % 'restart_httpd',
            }
        )
    else:
        logger.debug('HTTP cert not issued by IPA, \'%s\', skip tracking '
                     'check', DN(cert.issuer))

    # Check the ldap server cert if issued by IPA
    ds_nickname = ds.get_server_cert_nickname(serverid)
    ds_db_dirname = dsinstance.config_dirname(serverid)
    ds_db = certs.CertDB(api.env.realm, nssdir=ds_db_dirname)
    connected = api.Backend.ldap2.isconnected()
    if not connected:
        logger.debug('Unable to determine if \'%s\' was issued by IPA '
                     'because no LDAP connection, assuming yes.')
    if not connected or ds_db.is_ipa_issued_cert(api, ds_nickname):
        requests.append(
            {
                'cert-database': ds_db_dirname[:-1],
                'cert-nickname': ds_nickname,
                'ca-name': 'IPA',
                'cert-postsave-command':
                    '%s %s' % (template % 'restart_dirsrv', serverid),
            }
        )
    else:
        logger.debug('DS cert is not issued by IPA, skip tracking check')

    # Check if pkinit is enabled
    if os.path.exists(paths.KDC_CERT):
        pkinit_request_ca = krbinstance.get_pkinit_request_ca()
        requests.append(
            {
                'cert-file': paths.KDC_CERT,
                'key-file': paths.KDC_KEY,
                'ca-name': pkinit_request_ca,
                'cert-postsave-command':
                    template % 'renew_kdc_cert',
            }
        )
    else:
        logger.debug('No KDC pkinit certificate')

    # See if a host certificate was issued. This is only to
    # prevent a false-positive if one is indeed installed.
    local = {
        paths.IPA_NSSDB_DIR: 'Local IPA host',
        paths.NSS_DB_DIR: 'IPA Machine Certificate - %s' % socket.getfqdn(),
    }
    for db, nickname in local.items():
        nssdb = certdb.NSSDatabase(db)
        if nssdb.has_nickname(nickname):
            requests.append(
                {
                    'cert-database': db,
                    'cert-nickname': nickname,
                    'ca-name': 'IPA',
                }
            )

    return requests


def expected_token(token_name, certmonger_token):
    """The value is stored in two places, do some sanity checking"""
    if token_name != str(certmonger_token):
        logger.debug(
            "The IPA token %s doesn't match the certmonger token "
            "%s.", token_name, certmonger_token
        )
        return False

    return True


def get_token_password(hsm_enabled, token):
    if not hsm_enabled:
        return None

    with open(paths.PKI_TOMCAT_PASSWORD_CONF, "r") as passfile:
        contents = passfile.readlines()

    for line in contents:
        data = line.split('=', 1)
        if data[0] == 'hardware-' + token:
            return data[1]

    return None


def get_token_password_file(hsm_enabled, token):
    """The CA contains the list of HSM passwords, find ours"""
    pwdfile = None
    token_pw = get_token_password(hsm_enabled, token)

    if hsm_enabled and token_pw:
        pwdfile = ipautil.write_tmp_file(token_pw)

    return pwdfile


def get_dogtag_cert_password():
    """Return the NSSDB token password

       Will raise IOError if there is a problem reading the file.
    """
    ca_passwd = None
    token = 'internal'
    with open(paths.PKI_TOMCAT_PASSWORD_CONF, 'r') as f:
        for line in f:
            (tok, pin) = line.split('=', 1)
            if token == tok:
                ca_passwd = pin.strip()
                break

    return ca_passwd


@registry
class IPACertmongerExpirationCheck(IPAPlugin):
    """
    Collect the known/tracked certificates and check the validity

    This verifies only the information that certmonger has and uses
    to schedule renewal.

    This is to ensure something hasn't changed certmonger's view of
    the world.
    """
    @duration
    def check(self):
        cm = certmonger._certmonger()

        all_requests = cm.obj_if.get_requests()
        for req in all_requests:
            request = certmonger._cm_dbus_object(cm.bus, cm, req,
                                                 certmonger.DBUS_CM_REQUEST_IF,
                                                 certmonger.DBUS_CM_IF, True)
            id = request.prop_if.Get(certmonger.DBUS_CM_REQUEST_IF,
                                     'nickname')
            notafter = request.prop_if.Get(certmonger.DBUS_CM_REQUEST_IF,
                                           'not-valid-after')
            if notafter == 0:
                yield Result(self, constants.ERROR,
                             key=id,
                             msg='certmonger request id {key} does not have '
                                 'a not-valid-after date, assuming it '
                                 'has not been issued yet.')
                continue

            nafter = datetime.fromtimestamp(notafter, timezone.utc)
            now = datetime.now(timezone.utc)

            if now > nafter:
                yield Result(self, constants.ERROR,
                             key=id,
                             expiration_date=generalized_time(nafter),
                             msg='Request id {key} expired on '
                                 '{expiration_date}')
            else:
                delta = nafter - now
                diff = int(delta.total_seconds() / DAY)
                if diff < int(self.config.cert_expiration_days):
                    yield Result(self, constants.WARNING,
                                 key=id,
                                 expiration_date=generalized_time(nafter),
                                 days=diff,
                                 msg='Request id {key} expires in {days} '
                                     'days. certmonger should renew this '
                                     'automatically. Watch the status with '
                                     'getcert list -i {key}.')
                else:
                    yield Result(self, constants.SUCCESS,
                                 key=id)


@registry
class IPACertfileExpirationCheck(IPAPlugin):
    """
    Collect the known/tracked certificates and check file validity

    Look into the certificate file or NSS database to check the
    validity of the on-disk certificate.

    This is to ensure a certificate wasn't replaced without
    certmonger being notified.
    """
    @duration
    def check(self):
        cm = certmonger._certmonger()

        all_requests = cm.obj_if.get_requests()
        for req in all_requests:
            request = certmonger._cm_dbus_object(cm.bus, cm, req,
                                                 certmonger.DBUS_CM_REQUEST_IF,
                                                 certmonger.DBUS_CM_IF, True)
            id = request.prop_if.Get(certmonger.DBUS_CM_REQUEST_IF,
                                     'nickname')

            store = request.prop_if.Get(certmonger.DBUS_CM_REQUEST_IF,
                                        'cert-storage')
            if store == 'FILE':
                certfile = str(request.prop_if.Get(
                               certmonger.DBUS_CM_REQUEST_IF, 'cert-file'))
                try:
                    cert = x509.load_certificate_from_file(certfile)
                except Exception as e:
                    yield Result(self, constants.ERROR,
                                 key=id,
                                 certfile=certfile,
                                 error=str(e),
                                 msg='Request id {key}: Unable to open cert '
                                     'file \'{certfile}\': {error}')
                    continue
            elif store == 'NSSDB':
                token = request.prop_if.Get(certmonger.DBUS_CM_REQUEST_IF,
                                            'key-token')
                nickname = str(request.prop_if.Get(
                               certmonger.DBUS_CM_REQUEST_IF, 'key_nickname'))
                if token and expected_token(self.ca.token_name, token):
                    nickname = '{}:{}'.format(token, nickname)
                dbdir = str(request.prop_if.Get(
                            certmonger.DBUS_CM_REQUEST_IF, 'cert_database'))

                pwd_file = get_token_password_file(self.ca.hsm_enabled,
                                                   token)

                try:
                    if 'pwd_file' in signature(certdb.NSSDatabase).parameters:
                        # pylint: disable=unexpected-keyword-arg
                        db = certdb.NSSDatabase(
                            dbdir, token=token,
                            pwd_file=pwd_file.name if pwd_file else None)
                    else:
                        # Fall back to older API
                        db = certdb.NSSDatabase(dbdir)
                except Exception as e:
                    yield Result(self, constants.ERROR,
                                 key=id,
                                 dbdir=dbdir,
                                 error=str(e),
                                 msg='Request id {key}: Unable to open NSS '
                                     'database \'{dbdir}\': {error}')
                    continue

                try:
                    cert = db.get_cert(nickname)
                except Exception as e:
                    yield Result(self, constants.ERROR,
                                 key=id,
                                 dbdir=dbdir,
                                 nickname=nickname,
                                 error=str(e),
                                 msg='Request id {key}: Unable to retrieve '
                                     'cert \'{nickname}\' from \'{dbdir}\': '
                                     '{error}')
                    continue
            else:
                yield Result(self, constants.ERROR,
                             key=id,
                             store=store,
                             msg='Request id {key}: Unknown certmonger '
                                 'storage type: {store}')
                continue

            now = datetime.utcnow()
            notafter = cert.not_valid_after

            if now > notafter:
                yield Result(self, constants.ERROR,
                             key=id,
                             expiration_date=generalized_time(notafter),
                             msg='Request id {key} expired on '
                                 '{expiration_date}')
                continue

            delta = notafter - now
            diff = int(delta.total_seconds() / DAY)
            if diff < int(self.config.cert_expiration_days):
                yield Result(self, constants.WARNING,
                             key=id,
                             expiration_date=generalized_time(notafter),
                             days=diff,
                             msg='Request id {key} expires in {days} '
                                 'days. certmonger should renew this '
                                 'automatically. Watch the status with'
                                 'getcert list -i {key}.')
            else:
                yield Result(self, constants.SUCCESS, key=id)


@registry
class IPACertTracking(IPAPlugin):
    """Compare the certificates tracked by certmonger to those that
       are configured by default.

       Steps:
       1. Collect all expected certificates into `requests`
       2. Get the ids of all the certificates that certmonger is tracking
       3. Iterate over `requests` to retrieve the request id of the
          expected tracking.
       4. If the id is found we remove it from the ids list and move on
       5. In the unlikely event that the request_id is not in the
          ids list of all tracked certs report it.
       6. Report on all tracked certs that IPA didn't setup itself as
          potential issues.
    """

    requires = ('dirsrv',)

    @duration
    def check(self):
        requests = get_expected_requests(self.ca, self.ds, self.serverid)
        cm = certmonger._certmonger()

        ids = []
        all_requests = cm.obj_if.get_requests()
        for req in all_requests:
            request = certmonger._cm_dbus_object(cm.bus, cm, req,
                                                 certmonger.DBUS_CM_REQUEST_IF,
                                                 certmonger.DBUS_CM_IF, True)
            id = request.prop_if.Get(certmonger.DBUS_CM_REQUEST_IF,
                                     'nickname')
            ids.append(str(id))

        for request in requests:
            request_id = certmonger.get_request_id(request)
            try:
                if request_id is not None:
                    # Tracking found, move onto the next
                    ids.remove(request_id)
                    yield Result(self, constants.SUCCESS,
                                 key=request_id)
                    continue
            except ValueError as e:
                # A request was found but the id isn't in the
                # list from certmonger!?
                yield Result(self, constants.ERROR,
                             key=request_id,
                             error=str(e),
                             msg='Found request id {key} but it is not tracked'
                                 ' by certmonger!?: {error}')
                continue

            # The criteria was not met
            if request_id is None:
                flatten = ', '.join("{!s}={!s}".format(key, val)
                                    for (key, val) in request.items())
                yield Result(self, constants.ERROR,
                             key=flatten,
                             msg='Expected certmonger tracking is missing for '
                                 '{key}. Automated renewal will not happen '
                                 'for this certificate')
                continue

        # Report any unknown certmonger requests as warnings
        if ids:
            for id in ids:
                yield Result(self, constants.WARNING, key=id,
                             msg='certmonger tracking request {key} found and '
                                 'is not expected on an IPA master.')


@registry
class IPACertDNSSAN(IPAPlugin):
    """Check whether a IPA-issued certificates have a SAN configured

       Steps:
       1. Collect all expected certificates into `requests`
       2. Iterate over the list of certificates
       3. If issued by IPA and a caIPAserviceCert then verify that
          the host FQDN is in the list of SAN
       4. If a CA is configured on this host then also verify that
          ipa-ca.$DOMAIN is in the SAN.
    """

    requires = ('dirsrv',)

    @duration
    def check(self):
        fqdn = socket.getfqdn()
        requests = get_expected_requests(self.ca, self.ds, self.serverid)

        for request in requests:
            request_id = certmonger.get_request_id(request)
            if request_id is None:
                # log and skip. Missed tracking is reported by IPACertTracking
                flatten = ', '.join("{!s}={!s}".format(key, val)
                                    for (key, val) in request.items())

                logger.debug(
                    "Skipping %s since it is handled by IPACertTracking",
                    flatten
                )
                continue

            ca_name = certmonger.get_request_value(request_id, 'ca-name')
            if ca_name != 'IPA':
                logger.debug('Skipping request %s with CA %s',
                             request_id, ca_name)
                continue
            profile = certmonger.get_request_value(request_id,
                                                   'template_profile')
            if profile != 'caIPAserviceCert':
                logger.debug('Skipping request %s with profile %s',
                             request_id, profile)
                continue

            certfile = None
            if request.get('cert-file') is not None:
                certfile = request.get('cert-file')
                try:
                    cert = x509.load_certificate_from_file(certfile)
                except Exception as e:
                    yield Result(self, constants.ERROR,
                                 key=request_id,
                                 certfile=certfile,
                                 error=str(e),
                                 msg='Unable to open cert file {certfile}: '
                                     '{error}')
                    continue
            elif request.get('cert-database') is not None:
                nickname = request.get('cert-nickname')
                dbdir = request.get('cert-database')
                try:
                    db = certdb.NSSDatabase(dbdir)
                except Exception as e:
                    yield Result(self, constants.ERROR,
                                 key=request_id,
                                 dbdir=dbdir,
                                 error=str(e),
                                 msg='Unable to open NSS database {dbdir}: '
                                     '{error}')
                    continue
                try:
                    cert = db.get_cert(nickname)
                except Exception as e:
                    yield Result(self, constants.ERROR,
                                 key=id,
                                 dbdir=dbdir,
                                 nickname=nickname,
                                 error=str(e),
                                 msg='Unable to retrieve certificate '
                                     '\'{nickname}\' from {dbdir}: {error}')
                    continue

            hostlist = [fqdn]
            if self.ca.is_configured() and certfile == paths.HTTPD_CERT_FILE:
                hostlist.append(f'{IPA_CA_RECORD}.{api.env.domain}')
            error = False
            for host in hostlist:
                if host not in cert.san_a_label_dns_names:
                    error = True
                    yield Result(self, constants.ERROR,
                                 key=request_id,
                                 hostname=host,
                                 san=cert.san_a_label_dns_names,
                                 ca=ca_name,
                                 profile=profile,
                                 msg='Certificate request id {key} with '
                                     'profile {profile} for CA {ca} does not '
                                     'have a DNS SAN {san} matching name '
                                     '{hostname}')
            if not error:
                yield Result(self, constants.SUCCESS,
                             key=request_id,
                             hostname=hostlist,
                             san=cert.san_a_label_dns_names,
                             ca=ca_name,
                             profile=profile)


@registry
class IPACertNSSTrust(IPAPlugin):
    """Compare the NSS trust for the CA certs to a known good value"""

    @duration
    def check(self):
        if not self.ca.is_configured():
            logger.debug('CA is not configured, skipping NSS trust check')
            return

        if self.ca.hsm_enabled:
            token = self.ca.token_name + ":"
        else:
            token = ""

        expected_trust = {
            f'{token}ocspSigningCert cert-pki-ca': 'u,u,u',
            f'{token}subsystemCert cert-pki-ca': 'u,u,u',
            f'{token}auditSigningCert cert-pki-ca': 'u,u,Pu',
            'Server-Cert cert-pki-ca': 'u,u,u',
        }
        kra = krainstance.KRAInstance(api.env.realm)
        if kra.is_installed():
            kra_trust = {
                f'{token}transportCert cert-pki-kra': 'u,u,u',
                f'{token}storageCert cert-pki-kra': 'u,u,u',
                f'{token}auditSigningCert cert-pki-kra': 'u,u,Pu',
            }
            expected_trust.update(kra_trust)

        db = certdb.NSSDatabase(paths.PKI_TOMCAT_ALIAS_DIR)
        certlist = db.list_certs()
        if token:
            token = token[:-1]  # Strip off trailing colon

            pwd_file = get_token_password_file(self.ca.hsm_enabled,
                                               token)

            # pylint: disable=unexpected-keyword-arg
            db = certdb.NSSDatabase(
                paths.PKI_TOMCAT_ALIAS_DIR, token=token,
                pwd_file=pwd_file.name if pwd_file else None)

            certlist += db.list_certs()

        for nickname, _trust_flags in certlist:
            flags = certdb.unparse_trust_flags(_trust_flags)
            if nickname == f'{token}caSigningCert cert-pki-ca':
                expected = 'CTu,Cu,Cu'
            else:
                try:
                    expected = expected_trust[nickname]
                except KeyError:
                    logger.debug(
                        "%s not found in %s, assuming 3rd party",
                        nickname,
                        paths.PKI_TOMCAT_ALIAS_DIR,
                    )
                    continue
            try:
                expected_trust.pop(nickname)
            except KeyError:
                pass
            if flags != expected:
                yield Result(
                    self, constants.ERROR, key=nickname,
                    expected=expected,
                    got=flags,
                    nickname=nickname,
                    dbdir=paths.PKI_TOMCAT_ALIAS_DIR,
                    msg='Incorrect NSS trust for {nickname} in {dbdir}. '
                        'Got {got} expected {expected}.')
                continue

            yield Result(self, constants.SUCCESS, key=nickname)

        for nickname in expected_trust:
            yield Result(
                self, constants.ERROR,
                key=nickname,
                nickname=nickname,
                dbdir=paths.PKI_TOMCAT_ALIAS_DIR,
                msg='Certificate {nickname} missing from {dbdir} while '
                    'verifying trust')


@registry
class IPACertMatchCheck(IPAPlugin):
    """
    Ensure certificates match between LDAP and NSS databases
    """

    requires = ('dirsrv',)

    def get_cert_list_from_db(self, nssdb, nickname):
        """
        Retrieve all certificates from an NSS database for nickname.
        """
        try:
            args = ["-L", "-n", nickname, "-a"]
            result = nssdb.run_certutil(args, capture_output=True)
            return x509.load_certificate_list(result.raw_output)
        except ipautil.CalledProcessError:
            return []

    @duration
    def check(self):
        if not self.ca.is_configured():
            logger.debug("No CA configured, skipping certificate match check")
            return

        # Ensure /etc/ipa/ca.crt matches the NSS DB CA certificates
        def match_cacert_and_db(plugin, cacerts, dbpath):
            db = certs.CertDB(api.env.realm, dbpath)
            nickname = '%s IPA CA' % api.env.realm
            try:
                dbcacerts = self.get_cert_list_from_db(db, nickname)
            except Exception as e:
                yield Result(plugin, constants.ERROR,
                             key=nickname,
                             error=str(e),
                             msg='Unable to load CA cert: {error}')
                return False

            ok = True
            for cert in dbcacerts:
                if cert not in cacerts:
                    ok = False
                    yield Result(plugin, constants.ERROR,
                                 key=nickname,
                                 nickname=nickname,
                                 serial_number=cert.serial_number,
                                 dbdir=dbpath,
                                 certdir=paths.IPA_CA_CRT,
                                 msg=('CA Certificate nickname {nickname} '
                                      'with serial number {serial} '
                                      'is in {dbdir} but is not in'
                                      '%s' % paths.IPA_CA_CRT))
            return ok

        try:
            cacerts = x509.load_certificate_list_from_file(paths.IPA_CA_CRT)
        except Exception:
            yield Result(self, constants.ERROR,
                         key=paths.IPA_CA_CRT.replace(os.path.sep, '_'),
                         path=paths.IPA_CA_CRT,
                         msg='Unable to load CA cert file {path}: {error}')
            return

        # Ensure CA cert entry from LDAP matches /etc/ipa/ca.crt
        dn = DN('cn=%s IPA CA' % api.env.realm,
                'cn=certificates,cn=ipa,cn=etc',
                api.env.basedn)
        try:
            entry = self.conn.get_entry(dn)
        except errors.NotFound:
            yield Result(self, constants.ERROR,
                         key=str(dn),
                         dn=str(dn),
                         msg='CA Certificate entry \'{dn}\' '
                             'not found in LDAP')
            return

        cacerts_ok = True
        # Are all the certs in LDAP for the IPA CA in /etc/ipa/ca.crt
        for cert in entry['CACertificate']:
            if cert not in cacerts:
                cacerts_ok = False
                yield Result(self, constants.ERROR,
                             key=str(dn),
                             dn=str(dn),
                             serial_number=cert.serial_number,
                             msg=('CA Certificate serial number {serial} is '
                                  'in LDAP \'{dn}\' but is not in '
                                  '%s' % paths.IPA_CA_CRT))

        # Ensure NSS DBs have matching CA certs for /etc/ipa/ca.crt
        serverid = realm_to_serverid(api.env.realm)
        dspath = paths.ETC_DIRSRV_SLAPD_INSTANCE_TEMPLATE % serverid

        cacertds_ok = yield from match_cacert_and_db(self, cacerts, dspath)
        cacertnss_ok = yield from match_cacert_and_db(self, cacerts,
                                                      paths.IPA_NSSDB_DIR)
        if cacerts_ok:
            yield Result(self, constants.SUCCESS,
                         key=paths.IPA_CA_CRT)
        if cacertds_ok:
            yield Result(self, constants.SUCCESS,
                         key=dspath)
        if cacertnss_ok:
            yield Result(self, constants.SUCCESS,
                         key=paths.IPA_NSSDB_DIR)


@registry
class IPADogtagCertsMatchCheck(IPAPlugin):
    """
    Check if dogtag certs present in both NSS DB and LDAP match
    """
    requires = ('dirsrv',)

    @duration
    def check(self):
        if not self.ca.is_configured():
            logger.debug('CA is not configured, skipping match check')
            return

        def match_ldap_nss_cert(plugin, ldap, db, cert_dn, attr, cert_nick):
            try:
                entry = ldap.get_entry(cert_dn)
            except errors.NotFound:
                yield Result(plugin, constants.ERROR,
                             key=cert_dn,
                             msg='%s entry not found in LDAP' % cert_dn)
                return False
            try:
                nsscert = db.get_cert_from_db(cert_nick)
            except Exception as e:
                yield Result(plugin, constants.ERROR,
                             key=cert_nick,
                             error=str(e),
                             msg=('Unable to load %s certificate:'
                                  '{error}' % cert_nick))
                return False
            cert_matched = any(cert == nsscert for cert in entry[attr])
            if not cert_matched:
                yield Result(plugin, constants.ERROR,
                             key=cert_nick,
                             nickname=cert_nick,
                             dbdir=db.secdir,
                             msg=('{nickname} certificate in NSS DB {dbdir} '
                                  'does not match entry in LDAP'))
                return False
            return True

        def match_ldap_nss_certs_by_subject(plugin, ldap, db, dn,
                                            expected_nicks_subjects):
            all_ok = True
            for nick, subject in expected_nicks_subjects.items():
                entries = ldap.get_entries(
                    dn,
                    filter=f'subjectname={subject}'
                )
                cert = db.get_cert_from_db(nick)
                ok = any(
                    cert in entry["userCertificate"]
                    for entry in entries
                    if "userCertificate" in entry
                )
                if not ok:
                    all_ok = False
                    yield Result(plugin, constants.ERROR,
                                 key=nick,
                                 nickname=nick,
                                 dbdir=db.secdir,
                                 msg=('{nickname} certificate in NSS DB '
                                      '{dbdir} does not match entry in LDAP'))
            return all_ok

        if self.ca.hsm_enabled:
            token = self.ca.token_name + ':'
        else:
            token = ''
        pwd_file = get_token_password_file(self.ca.hsm_enabled,
                                           self.ca.token_name)
        if 'pwd_file' in signature(certs.CertDB).parameters:
            # pylint: disable=unexpected-keyword-arg
            db = certs.CertDB(api.env.realm, paths.PKI_TOMCAT_ALIAS_DIR,
                              pwd_file=pwd_file.name if pwd_file else None)
        else:
            # Fall back to older API
            db = certs.CertDB(api.env.realm, paths.PKI_TOMCAT_ALIAS_DIR)
        dn = DN('uid=pkidbuser,ou=people,o=ipaca')
        subsystem_nick = f'{token}subsystemCert cert-pki-ca'
        subsystem_ok = yield from match_ldap_nss_cert(self, self.conn,
                                                      db, dn,
                                                      'userCertificate',
                                                      subsystem_nick)
        dn = DN('cn=%s IPA CA' % api.env.realm,
                'cn=certificates,cn=ipa,cn=etc',
                api.env.basedn)
        casigning_nick = f'{token}caSigningCert cert-pki-ca'
        casigning_ok = yield from match_ldap_nss_cert(self, self.conn,
                                                      db, dn, 'CACertificate',
                                                      casigning_nick)

        config = api.Command.config_show()
        subject_base = config['result']['ipacertificatesubjectbase'][0]
        expected_nicks_subjects = {
            f'{token}ocspSigningCert cert-pki-ca':
                f'CN=OCSP Subsystem,{subject_base}',
            f'{token}subsystemCert cert-pki-ca':
                f'CN=CA Subsystem,{subject_base}',
            f'{token}auditSigningCert cert-pki-ca':
                f'CN=CA Audit,{subject_base}',
            'Server-Cert cert-pki-ca':
                f'CN={api.env.host},{subject_base}',
        }

        kra = krainstance.KRAInstance(api.env.realm)
        if kra.is_installed():
            kra_expected_nicks_subjects = {
                f'{token}transportCert cert-pki-kra':
                    f'CN=KRA Transport Certificate,{subject_base}',
                f'{token}storageCert cert-pki-kra':
                    f'CN=KRA Storage Certificate,{subject_base}',
                f'{token}auditSigningCert cert-pki-kra':
                    f'CN=KRA Audit,{subject_base}',
            }
            expected_nicks_subjects.update(kra_expected_nicks_subjects)

        ipaca_basedn = DN('ou=certificateRepository,ou=ca,o=ipaca')
        ipaca_certs_ok = yield from match_ldap_nss_certs_by_subject(
                                    self, self.conn, db,
                                    ipaca_basedn,
                                    expected_nicks_subjects
                                )

        if subsystem_ok:
            yield Result(self, constants.SUCCESS,
                         key=subsystem_nick)
        if casigning_ok:
            yield Result(self, constants.SUCCESS,
                         key=casigning_nick)
        if ipaca_certs_ok:
            yield Result(self, constants.SUCCESS,
                         key=str(ipaca_basedn))


@registry
class IPANSSChainValidation(IPAPlugin):
    """Validate the certificate chain of the certs."""

    def validate_nss(self, dbdir, dbtype, pinfile, nickname):
        """Call out to certutil to verify a certificate.

           The caller must handle the exceptions
        """
        args = [paths.CERTUTIL, '-V', '-u', 'V', '-e']
        args.extend(['-d', dbtype + ':' + dbdir])
        args.extend(['-n', nickname])
        args.extend(['-f', pinfile])

        return ipautil.run(args, raiseonerr=False)

    @duration
    def check(self):
        validate = []
        ca_pw_fname = None

        if self.ca.is_configured():
            try:
                ca_passwd = get_dogtag_cert_password()
            except IOError as e:
                yield Result(
                    self, constants.ERROR,
                    key='db_authenticate',
                    error=str(e),
                    msg='Unable to read CA NSSDB token password: {error}')
                return
            else:
                with tempfile.NamedTemporaryFile(mode='w',
                                                 delete=False) as ca_pw_file:
                    ca_pw_file.write(ca_passwd)
                    ca_pw_fname = ca_pw_file.name

                validate.append(
                    (
                        paths.PKI_TOMCAT_ALIAS_DIR,
                        'Server-Cert cert-pki-ca',
                        ca_pw_fname,
                    ),
                )

        validate.append(
            (
                dsinstance.config_dirname(self.serverid),
                self.ds.get_server_cert_nickname(self.serverid),
                os.path.join(dsinstance.config_dirname(self.serverid),
                             'pwdfile.txt'),
            )
        )

        # Wrap in try/except to ensure the temporary password file is
        # removed
        try:
            for (dbdir, nickname, pinfile) in validate:
                # detect the database type so we have the right prefix
                db = certdb.NSSDatabase(dbdir)

                key = os.path.normpath(dbdir) + ':' + nickname
                try:
                    response = self.validate_nss(dbdir, db.dbtype, pinfile,
                                                 nickname)
                except ipautil.CalledProcessError as e:
                    logger.debug('Validation of NSS certificate failed %s', e)
                    yield Result(
                        self, constants.ERROR,
                        key=key,
                        dbdir=dbdir,
                        nickname=nickname,
                        reason=response.output_error,
                        msg='Validation of {nickname} in {dbdir} failed: '
                            '{reason}')
                else:
                    if 'certificate is valid' not in \
                            response.raw_output.decode('utf-8'):
                        yield Result(
                            self, constants.ERROR,
                            key=key,
                            dbdir=dbdir,
                            nickname=nickname,
                            reason="%s: %s" %
                            (response.raw_output.decode('utf-8'),
                             response.error_log),
                            msg='Validation of {nickname} in {dbdir} failed: '
                                '{reason}')
                    else:
                        yield Result(self, constants.SUCCESS,
                                     dbdir=dbdir, nickname=nickname,
                                     key=key)
        finally:
            if ca_pw_fname:
                ipautil.remove_file(ca_pw_fname)


@registry
class IPAOpenSSLChainValidation(IPAPlugin):
    """Validate the certificate chain of the certs."""

    def validate_openssl(self, file):
        """Call out to openssl to verify a certificate against global chain

           The caller must handle the exceptions
        """
        args = [paths.OPENSSL, 'verify',
                '-verbose',
                '-show_chain',
                '-CAfile', paths.IPA_CA_CRT,
                file]

        return ipautil.run(args, raiseonerr=False)

    @duration
    def check(self):
        certlist = [paths.HTTPD_CERT_FILE]
        if self.ca.is_configured():
            certlist.append(paths.RA_AGENT_PEM)

        for cert in certlist:
            try:
                response = self.validate_openssl(cert)
            except Exception as e:
                yield Result(
                    self, constants.ERROR,
                    key=cert,
                    error=str(e),
                    msg='Certificate validation for {key} failed: {error}')
                continue
            else:
                if ': OK' not in response.raw_output.decode('utf-8'):
                    yield Result(
                        self, constants.ERROR, key=cert,
                        reason=response.raw_error_output.decode('utf-8'),
                        msg='Certificate validation for {key} failed: '
                            '{reason}')
                else:
                    yield Result(
                        self, constants.SUCCESS, key=cert)


def check_agent(plugin, base_dn, agent_type):
    """Check RA/KRA Agent"""

    try:
        cert = x509.load_certificate_from_file(paths.RA_AGENT_PEM)
    except Exception as e:
        yield Result(plugin, constants.ERROR,
                     key=paths.RA_AGENT_PEM.replace(os.path.sep, '_'),
                     error=str(e),
                     msg='Unable to load RA cert: {error}')
        return
    serial_number = cert.serial_number
    subject = DN(cert.subject)
    issuer = DN(cert.issuer)
    description = '2;%d;%s;%s' % (serial_number, issuer, subject)
    logger.debug('%s agent description should be %s', agent_type, description)
    db_filter = ldap2.ldap2.combine_filters(
        [
            ldap2.ldap2.make_filter({'objectClass': 'inetOrgPerson'}),
            ldap2.ldap2.make_filter(
                {'description': ';%s;%s' % (issuer, subject)},
                exact=False, trailing_wildcard=False),
        ],
        ldap2.ldap2.MATCH_ALL)
    try:
        entries = plugin.conn.get_entries(base_dn,
                                          plugin.conn.SCOPE_SUBTREE,
                                          db_filter)
    except errors.NotFound:
        yield Result(plugin, constants.ERROR,
                     key=agent_type,
                     description=description,
                     msg='%s agent not found in LDAP' % agent_type)
        return
    except Exception as e:
        yield Result(plugin, constants.ERROR,
                     key=agent_type,
                     error=str(e),
                     msg='Retrieving %s agent from LDAP failed {error}'
                         % agent_type)
        return
    else:
        logger.debug('%s agent description is %s', agent_type, description)
        if len(entries) != 1:
            yield Result(plugin, constants.ERROR,
                         key='too_many_agents',
                         found=len(entries),
                         msg='Too many %s agent entries found, {found}'
                             % agent_type)
            return
        entry = entries[0]
        raw_desc = entry.get('description')
        if raw_desc is None:
            yield Result(plugin, constants.ERROR,
                         key='agent_missing_description',
                         msg='%s agent is missing the description '
                             'attribute or it is not readable' % agent_type)
            return
        ra_desc = raw_desc[0]
        ra_certs = entry.get('usercertificate')
        (_version, exp_serial, exp_issuer, exp_subject) = \
            ra_desc.split(';')
        matched = all(
            [
                str(serial_number) == exp_serial,
                DN(issuer) == DN(exp_issuer),
                DN(subject) == DN(exp_subject),
            ]
        )
        if not matched:
            yield Result(plugin, constants.ERROR,
                         key='description_mismatch',
                         expected=description,
                         got=ra_desc,
                         msg='%s agent description does not match. Found '
                         '{got} in LDAP and expected {expected}' % agent_type)
            return
        found = False
        for candidate in ra_certs:
            if candidate == cert:
                found = True
                break
        if not found:
            yield Result(plugin, constants.ERROR,
                         key='ldap_mismatch',
                         certfile=paths.RA_AGENT_PEM,
                         dn=str(entry.dn),
                         msg='%s agent certificate in {certfile} not '
                             'found in LDAP userCertificate attribute '
                             'for the entry {dn}' % agent_type)
        yield Result(plugin, constants.SUCCESS)


@registry
class IPARAAgent(IPAPlugin):
    """Validate the RA Agent used to talk to the CA

       Compare the description and usercertificate values.
    """

    requires = ('dirsrv',)

    @duration
    def check(self):
        if not self.ca.is_configured():
            logger.debug('CA is not configured, skipping RA Agent check')
            return

        base_dn = DN('uid=ipara,ou=people,o=ipaca')
        yield from check_agent(self, base_dn, 'RA')


@registry
class IPAKRAAgent(IPAPlugin):
    """Validate the KRA Agent

       Compare the description and usercertificate values.
    """

    requires = ('dirsrv',)

    @duration
    def check(self):
        if not self.ca.is_configured():
            logger.debug('CA is not configured, skipping KRA Agent check')
            return

        kra = krainstance.KRAInstance(api.env.realm)
        if not kra.is_installed():
            logger.debug('KRA is not installed, skipping KRA Agent check')
            return

        base_dn = DN('uid=ipakra,ou=people,o=kra,o=ipaca')
        yield from check_agent(self, base_dn, 'KRA')


@registry
class IPACertRevocation(IPAPlugin):
    """Confirm that the IPA certificates are not revoked

       This uses the certmonger expected tracking list to know which
       one(s) to consider.
    """

    revocation_reason = [
        "unspecified",
        "key compromise",
        "CA compromise",
        "affiliation changed",
        "superseded",
        "cessation of operation",
        "certificate hold",
        "",  # unused
        "remove from CRL",
        "privilege withdrawn",
        "AA compromise",
    ]

    requires = ('dirsrv',)

    @duration
    def check(self):
        # For simplicity use the expected certmonger tracking for the
        # list of certificates to check because it already filters out
        # based on whether the CA system is configure and whether the
        # certificates were issued by IPA.
        if not self.ca.is_configured():
            logger.debug('CA is not configured, skipping revocation check')
            return
        requests = get_expected_requests(self.ca, self.ds, self.serverid)
        pwd_file = get_token_password_file(self.ca.hsm_enabled,
                                           self.ca.token_name)
        for request in requests:
            id = certmonger.get_request_id(request)
            if request.get('cert-file') is not None:
                certfile = request.get('cert-file')
                try:
                    cert = x509.load_certificate_from_file(certfile)
                except Exception as e:
                    yield Result(self, constants.ERROR,
                                 key=id,
                                 certfile=certfile,
                                 error=str(e),
                                 msg='Unable to open cert file {certfile}: '
                                     '{error}')
                    continue
            elif request.get('cert-database') is not None:
                nickname = request.get('cert-nickname')
                token = request.get('key-token')
                if token == 'internal':
                    token = None
                dbdir = request.get('cert-database')
                try:
                    if 'pwd_file' in signature(certdb.NSSDatabase).parameters:
                        # pylint: disable=unexpected-keyword-arg
                        db = certdb.NSSDatabase(
                            dbdir, token=token,
                            pwd_file=pwd_file.name if pwd_file else None
                        )
                    else:
                        # Fall back to older API that doesn't support tokens
                        db = certdb.NSSDatabase(dbdir)
                except Exception as e:
                    yield Result(self, constants.ERROR,
                                 key=id,
                                 dbdir=dbdir,
                                 error=str(e),
                                 msg='Unable to open NSS database {dbdir}: '
                                     '{error}')
                    continue
                if token:
                    nickname = f'{token}:{nickname}'
                try:
                    cert = db.get_cert(nickname)
                except Exception as e:
                    yield Result(self, constants.ERROR,
                                 key=id,
                                 dbdir=dbdir,
                                 nickname=nickname,
                                 error=str(e),
                                 msg='Unable to retrieve certificate '
                                     '\'{nickname}\' from {dbdir}: {error}')
                    continue
            else:
                yield Result(self, constants.ERROR,
                             key=id,
                             msg='Unable to to identify certificate storage '
                                 'type for request {key}')
                continue

            issued = is_ipa_issued_cert(api, cert)
            if issued is False:
                logger.debug('\'%s\' was not issued by IPA, skipping',
                             DN(cert.subject))
                continue
            if issued is None:
                logger.debug('LDAP is down, skipping \'%s\'',
                             DN(cert.subject))
                continue

            # Now we have the cert either way, check the recovation
            try:
                result = api.Command.cert_show(cert.serial_number,
                                               all=True)
            except Exception as e:
                yield Result(self, constants.ERROR,
                             key=id,
                             serial=cert.serial_number,
                             error=str(e),
                             msg='Request for certificate serial number '
                                 '{serial} in request {key} failed: {error}')
                continue

            try:
                if result['result']['revoked']:
                    reason = result['result']['revocation_reason']
                    reason_txt = self.revocation_reason[reason]
                    yield Result(self, constants.ERROR,
                                 revocation_reason=reason_txt,
                                 key=id,
                                 msg='Certificate tracked by {key} is revoked '
                                     '{revocation_reason}')
                else:
                    yield Result(self, constants.SUCCESS, key=id)
            except Exception as e:
                yield Result(self, constants.ERROR,
                             key=id,
                             error=str(e),
                             msg='Unable to determine revocation '
                                 'status for {key}: {error}')


@registry
class IPACertmongerCA(IPAPlugin):
    """Ensure that the required CAs are available in certmonger

       Addresses symptom of https://pagure.io/freeipa/issue/7870
    """

    def find_ca(self, name):
        cm = certmonger._certmonger()
        ca_path = cm.obj_if.find_ca_by_nickname(name)
        return certmonger._cm_dbus_object(cm.bus, cm, ca_path,
                                          certmonger.DBUS_CM_CA_IF,
                                          certmonger.DBUS_CM_IF, True)

    @duration
    def check(self):
        ca_list = ['IPA']
        if self.ca.is_configured():
            ca_list.extend([
               'dogtag-ipa-ca-renew-agent',
               'dogtag-ipa-ca-renew-agent-reuse'
            ])
        for ca in ca_list:
            logger.debug('Checking for existence of certmonger CA \'%s\'',
                         ca)
            try:
                self.find_ca(ca)
            except Exception as e:
                logger.debug('Search for certmonger CA %s failed: %s', ca, e)
                yield Result(self, constants.ERROR,
                             key=ca,
                             msg='Certmonger CA \'{key}\' missing')
            else:
                yield Result(self, constants.SUCCESS,
                             key=ca)


@registry
class IPACAChainExpirationCheck(IPAPlugin):
    """Verify that the certs in the CA chain in /etc/ipa/ca.crt are valid
    """

    @duration
    def check(self):
        try:
            ca_certs = x509.load_certificate_list_from_file(paths.IPA_CA_CRT)
        except IOError as e:
            logger.debug("Could not open %s: %s", paths.IPA_CA_CRT, e)
            yield Result(self, constants.ERROR,
                         key=paths.IPA_CA_CRT,
                         error=str(e),
                         msg='Error opening IPA CA chain at {key}: {error}')
            return
        except ValueError as e:
            logger.debug(
                "% contains an invalid certificate", paths.IPA_CA_CRT
            )
            yield Result(self, constants.ERROR,
                         key=paths.IPA_CA_CRT,
                         error=str(e),
                         msg='IPA CA chain {key} contains an invalid '
                             'certificate: {error}')
            return

        now = datetime.now(timezone.utc)
        soon = now + timedelta(days=int(self.config.cert_expiration_days))
        for cert in ca_certs:
            subject = DN(cert.subject)
            subject = str(subject).replace('\\;', '\\3b')
            dt = cert.not_valid_after.replace(tzinfo=timezone.utc)
            if dt < now:
                logger.debug("%s is expired", subject)
                yield Result(self, constants.CRITICAL,
                             path=paths.IPA_CA_CRT,
                             key=subject,
                             msg='CA \'{key}\' in {path} is expired.')
            elif dt <= soon:
                logger.debug("%s is expiring soon", subject)
                yield Result(self, constants.WARNING,
                             path=paths.IPA_CA_CRT,
                             key=subject,
                             days=(dt - now).days,
                             msg='CA \'{key}\' in {path} is expiring in '
                                 '{days} days.')
            else:
                yield Result(self, constants.SUCCESS,
                             path=paths.IPA_CA_CRT,
                             key=subject,
                             days=(dt - now).days)



#
# Copyright (C) 2019 FreeIPA Contributors see COPYING for license
#
from unittest import mock, TestCase
from util import no_exceptions
from util import ADtrustBasedRole, ServiceBasedRole


class BaseTest(TestCase):
    """
    Base class for tests.

    Most tests use the same set of mocks so centralize and apply them
    once when the class of tests is created.

    A child class defines self.patches as a dictionary of functions
    and Mock values. These are applied once when the class starts up.

    If a test needs a particular value then it will need to use
    @patch individually.

    A default set of Mock patches is set because they apply to all or
    nearly all test cases.
    """
    default_patches = {
        'ipaserver.install.installutils.check_server_configuration':
        mock.Mock(return_value=None),
        'ipaserver.servroles.ServiceBasedRole':
        mock.Mock(return_value=ServiceBasedRole()),
        'ipaserver.servroles.ADtrustBasedRole':
        mock.Mock(return_value=ADtrustBasedRole()),
    }
    patches = {}
    results = None
    applied_patches = None

    def setup_class(self):
        # collect the list of patches to be applied for this class of
        # tests
        self.default_patches.update(self.patches)

        self.applied_patches = [
            mock.patch(patch, data) for patch, data in
            self.default_patches.items()
        ]

        for patch in self.applied_patches:
            patch.start()

    def teardown_class(self):
        mock.patch.stopall()

    def tearDown(self):
        """
        Ensure that no exceptions snuck into the results which might not
        be noticed because an exception may have the same result as
        the expected result.
        """
        no_exceptions(self.results)



#
# Copyright (C) 2019 FreeIPA Contributors see COPYING for license
#

from base import BaseTest
from unittest.mock import patch
from util import capture_results

from ipahealthcheck.core import config, constants
from ipahealthcheck.ipa.plugin import registry
from ipahealthcheck.ipa.dna import IPADNARangeCheck


class mock_ReplicationManager:
    def __init__(self, realm=None, host=None, start=None, max=None,
                 next=None, next_max=None):
        self.start = start
        self.max = max
        self.next = next
        self.next_max = next_max

    def get_DNA_range(self, host):
        return self.start, self.max

    def get_DNA_next_range(self, host):
        return self.next, self.next_max


class TestDNARange(BaseTest):
    @patch('ipaserver.install.replication.ReplicationManager')
    def test_dnarange_set(self, mock_manager):
        mock_manager.return_value = mock_ReplicationManager(start=1, max=100)
        framework = object()
        registry.initialize(framework, config.Config)
        f = IPADNARangeCheck(registry)

        self.results = capture_results(f)

        assert len(self.results) == 1

        result = self.results.results[0]
        assert result.result == constants.SUCCESS
        assert result.source == 'ipahealthcheck.ipa.dna'
        assert result.check == 'IPADNARangeCheck'
        assert result.kw.get('range_start') == 1
        assert result.kw.get('range_max') == 100
        assert result.kw.get('next_start') == 0
        assert result.kw.get('next_max') == 0

    @patch('ipaserver.install.replication.ReplicationManager')
    def test_dnarange_noset(self, mock_manager):
        mock_manager.return_value = mock_ReplicationManager()
        framework = object()
        registry.initialize(framework, config.Config)
        f = IPADNARangeCheck(registry)

        self.results = capture_results(f)

        assert len(self.results) == 1

        result = self.results.results[0]
        assert result.result == constants.WARNING
        assert result.source == 'ipahealthcheck.ipa.dna'
        assert result.check == 'IPADNARangeCheck'
        assert result.kw.get('range_start') == 0
        assert result.kw.get('range_max') == 0
        assert result.kw.get('next_start') == 0
        assert result.kw.get('next_max') == 0

    @patch('ipaserver.install.replication.ReplicationManager')
    def test_dnarange_next(self, mock_manager):
        mock_manager.return_value = mock_ReplicationManager(start=1,
                                                            max=100,
                                                            next=101,
                                                            next_max=200)
        framework = object()
        registry.initialize(framework, config.Config)
        f = IPADNARangeCheck(registry)

        self.results = capture_results(f)

        assert len(self.results) == 1

        result = self.results.results[0]
        assert result.result == constants.SUCCESS
        assert result.source == 'ipahealthcheck.ipa.dna'
        assert result.check == 'IPADNARangeCheck'
        assert result.kw.get('range_start') == 1
        assert result.kw.get('range_max') == 100
        assert result.kw.get('next_start') == 101
        assert result.kw.get('next_max') == 200



#
# Copyright (C) 2019 FreeIPA Contributors see COPYING for license
#

from util import capture_results, CAInstance
from base import BaseTest
from ipahealthcheck.core import constants, config
from ipahealthcheck.ipa.plugin import registry
from ipahealthcheck.ipa.certs import IPACertmongerCA
from unittest.mock import Mock, patch


class TestCertmonger(BaseTest):
    patches = {
        'ipaserver.install.cainstance.CAInstance':
        Mock(return_value=CAInstance()),
    }

    @patch('ipahealthcheck.ipa.certs.IPACertmongerCA.find_ca')
    def test_certmogner_ok(self, mock_find_ca):
        mock_find_ca.side_effect = [
            'IPA',
            'dogtag-ipa-ca-renew-agent',
            'dogtag-ipa-ca-renew-agent-reuse'
        ]
        framework = object()
        registry.initialize(framework, config.Config)
        f = IPACertmongerCA(registry)

        self.results = capture_results(f)

        assert len(self.results) == 3

        for result in self.results.results:
            assert result.result == constants.SUCCESS
            assert result.source == 'ipahealthcheck.ipa.certs'
            assert result.check == 'IPACertmongerCA'

    @patch('ipahealthcheck.ipa.certs.IPACertmongerCA.find_ca')
    def test_certmogner_missing(self, mock_find_ca):
        mock_find_ca.side_effect = [
            'IPA',
            'dogtag-ipa-ca-renew-agent',
        ]

        framework = object()
        registry.initialize(framework, config.Config)
        f = IPACertmongerCA(registry)

        self.results = capture_results(f)

        assert len(self.results) == 3

        for r in range(0, 1):
            result = self.results.results[r]
            assert result.result == constants.SUCCESS
            assert result.source == 'ipahealthcheck.ipa.certs'
            assert result.check == 'IPACertmongerCA'

        assert self.results.results[2].result == constants.ERROR
        assert self.results.results[2].kw.get('key') == \
            'dogtag-ipa-ca-renew-agent-reuse'



#
# Copyright (C) 2021 FreeIPA Contributors see COPYING for license
#

from base import BaseTest
from collections import namedtuple
from unittest.mock import patch
from util import capture_results

from ipaplatform.constants import constants as platform_constants

from ipahealthcheck.core import config, constants
from ipahealthcheck.ipa.plugin import registry
from ipahealthcheck.ipa.nss import IPAGroupMemberCheck


struct_group = namedtuple(
    'struct_group', ['gr_name', 'gr_passwd', 'gr_gid', 'gr_mem']
)


def make_group(name, members):
    return struct_group(name, 'x', 999, members)


class TestGroupMember(BaseTest):
    @patch('ipahealthcheck.ipa.nss.grp.getgrnam')
    def test_ipaapi_group_ok(self, mock_grp):
        mock_grp.return_value = make_group(
            platform_constants.IPAAPI_GROUP, (platform_constants.HTTPD_USER,),
        )
        framework = object()
        registry.initialize(framework, config.Config)
        registry.trust_agent = True
        f = IPAGroupMemberCheck(registry)

        self.results = capture_results(f)

        assert len(self.results) == 1

        result = self.results.results[0]
        assert result.result == constants.SUCCESS

    @patch('ipahealthcheck.ipa.nss.grp.getgrnam')
    def test_ipaapi_bad_group(self, mock_grp):
        mock_grp.side_effect = KeyError(
            f"name not found: '{platform_constants.IPAAPI_GROUP}'"
        )

        framework = object()
        registry.initialize(framework, config.Config)
        registry.trust_agent = True
        f = IPAGroupMemberCheck(registry)

        self.results = capture_results(f)

        assert len(self.results) == 1

        result = self.results.results[0]
        assert result.result == constants.ERROR
        assert result.kw.get('key') == platform_constants.IPAAPI_GROUP
        assert result.kw.get('msg') == 'group {key} does not exist'

    @patch('ipahealthcheck.ipa.nss.grp.getgrnam')
    def test_ipaapi_missing_member(self, mock_grp):
        mock_grp.return_value = make_group(
            platform_constants.IPAAPI_GROUP, ('foo',)
        )

        framework = object()
        registry.initialize(framework, config.Config)
        registry.trust_agent = True
        f = IPAGroupMemberCheck(registry)

        self.results = capture_results(f)

        assert len(self.results) == 1

        result = self.results.results[0]
        assert result.result == constants.ERROR
        assert result.kw.get('key') == platform_constants.IPAAPI_GROUP
        assert result.kw.get('msg') == \
            '{member} is not a member of group {key}'
